import functools
import os.path
import re
import aioipfs

from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QAction
from PyQt5.QtWidgets import QActionGroup
from PyQt5.QtWidgets import QMenu
from PyQt5.QtWidgets import QInputDialog
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QTextBrowser
from PyQt5.QtWidgets import QLineEdit

from PyQt5.QtPrintSupport import *

from PyQt5.QtCore import QUrl
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtCore import QTimer

from PyQt5 import QtWebEngineWidgets

from PyQt5.QtWebEngineWidgets import QWebEnginePage
from PyQt5.QtWebEngineWidgets import QWebEngineDownloadItem
from PyQt5.QtWebEngineWidgets import QWebEngineSettings
from PyQt5.QtWebEngineWidgets import QWebEngineContextMenuData
from PyQt5.QtWebChannel import QWebChannel

from PyQt5.QtWebEngineCore import QWebEngineUrlRequestInterceptor

from PyQt5.QtGui import QKeySequence

from galacteek import log
from galacteek import logUser
from galacteek import ensure

from galacteek.ipfs.wrappers import *
from galacteek.ipfs import cidhelpers
from galacteek.ipfs import megabytes
from galacteek.ipfs.stat import StatInfo
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.cidhelpers import joinIpfs
from galacteek.ipfs.cidhelpers import joinIpns
from galacteek.ipfs.cidhelpers import cidValid
from galacteek.core.asynclib import asyncify

from galacteek.core.schemes import isSchemeRegistered
from galacteek.core.schemes import SCHEME_ENS
from galacteek.core.schemes import SCHEME_IPFS
from galacteek.core.schemes import SCHEME_IPNS
from galacteek.core.schemes import SCHEME_DWEB
from galacteek.core.schemes import SCHEME_Z
from galacteek.core.schemes import DAGProxySchemeHandler
from galacteek.core.schemes import MultiDAGProxySchemeHandler
from galacteek.core.schemes import IPFSObjectProxyScheme

from galacteek.core.webprofiles import WP_NAME_IPFS
from galacteek.core.webprofiles import WP_NAME_MINIMAL
from galacteek.core.webprofiles import WP_NAME_WEB3

from galacteek.dweb.webscripts import scriptFromString
from galacteek.dweb.page import BaseHandler
from galacteek.dweb.page import pyqtSlot
from galacteek.dweb.render import renderTemplate
from galacteek.dweb.htmlparsers import IPFSLinksParser

from . import ui_browsertab
from .dag import DAGViewer
from .pin import PinBatchWidget, PinBatchTab
from .helpers import *
from .dialogs import *
from .hashmarks import *
from .i18n import *
from .clipboard import iCopyPathToClipboard
from .clipboard import iClipboardEmpty
from .history import HistoryMatchesWidget
from .widgets import *
from ..appsettings import *

# i18n


def iOpenInTab():
    return QCoreApplication.translate('BrowserTabForm', 'Open link in new tab')


def iOpenHttpInTab():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Open http/https link in new tab')


def iOpenLinkInTab():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Open link in new tab')


def iOpenWith():
    return QCoreApplication.translate('BrowserTabForm', 'Open with')


def iDownload():
    return QCoreApplication.translate('BrowserTabForm', 'Download')


def iSaveContainedWebPage():
    return QCoreApplication.translate(
        'BrowserTabForm', 'Save full webpage to the "Web Pages" folder')


def iJsConsole():
    return QCoreApplication.translate('BrowserTabForm', 'Javascript console')


def iPin():
    return QCoreApplication.translate('BrowserTabForm', 'PIN')


def iPinThisPage():
    return QCoreApplication.translate('BrowserTabForm', 'PIN (this page)')


def iPinRecursive():
    return QCoreApplication.translate('BrowserTabForm', 'PIN (recursive)')


def iSaveSelectedText():
    return QCoreApplication.translate(
        'BrowserTabForm', 'Save selected text to IPFS')


def iLinkToQaToolbar():
    return QCoreApplication.translate(
        'BrowserTabForm', 'Link to Quick Access toolbar')


def iFollow():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Follow IPNS resource')


def iEnterIpfsCID():
    return QCoreApplication.translate('BrowserTabForm', 'Enter an IPFS CID')


def iBrowseHomePage():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Go to home page')


def iBrowseIpfsCID():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Browse IPFS resource (CID)')


def iBrowseIpfsMultipleCID():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Browse multiple IPFS resources (CID)')


def iEnterIpfsCIDDialog():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Load IPFS CID dialog')


def iFollowIpnsDialog():
    return QCoreApplication.translate('BrowserTabForm',
                                      'IPNS add feed dialog')


def iBrowseIpnsHash():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Browse IPNS resource from hash/name')


def iBrowseCurrentClipItem():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Browse current clipboard item')


def iEnterIpns():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Enter an IPNS hash/name')


def iEnterIpnsDialog():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Load IPNS key dialog')


def iCreateQaMapping():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Create quick-access mapping')


def iHashmarked(path):
    return QCoreApplication.translate('BrowserTabForm',
                                      'Hashmarked {0}').format(path)


def iHashmarkTitleDialog():
    return QCoreApplication.translate('BrowserTabForm',
                                      'Hashmark title')


def iInvalidUrl(text):
    return QCoreApplication.translate('BrowserTabForm',
                                      'Invalid URL: {0}').format(text)


def iInvalidObjectPath(text):
    return QCoreApplication.translate(
        'BrowserTabForm',
        'Invalid IPFS object path: {0}').format(text)


def iInvalidCID(text):
    return QCoreApplication.translate(
        'BrowserTabForm',
        '{0} is an invalid IPFS CID (Content IDentifier)').format(text)


def iNotAnIpfsResource():
    return QCoreApplication.translate(
        'BrowserTabForm',
        'Not an IPFS resource')


def iWebProfileMinimal():
    return QCoreApplication.translate(
        'BrowserTabForm',
        'Minimal profile')


def iWebProfileIpfs():
    return QCoreApplication.translate(
        'BrowserTabForm',
        'IPFS profile')


def iWebProfileWeb3():
    return QCoreApplication.translate(
        'BrowserTabForm',
        'Web3 profile')


def iCidTooltipMessage(icon, rootCidV, more, rootCid, thisCid):
    return QCoreApplication.translate('BrowserTabForm',
                                      '''
    <p>
      <img src='{0}' width='32' height='32'/>
    </p>

    <p>Root CID: CIDv{1} {2} <b>{3}</b></p>

    <p>This page's CID: <b>{4}</b></p>

    <div style='font-size: 200%;'>
    <p>
      <b>Click on the cube</b> to get a view of the DAG for this page
    </p>
    </div>
''').format(icon, rootCidV, more, rootCid, thisCid)


def iIpnsTooltipMessage(icon, ipnsKey):
    return QCoreApplication.translate('BrowserTabForm',
                                      '''
        <p>
          <img src='{0}' width='32' height='32'/>
        </p>

        <p>IPNS domain/key <b>{1}</b></p>

        <div style='font-size: 200%;'>
        <p>
          <b>Click on the cube</b> to get a view of the DAG for this page
        </p>
        </div>
    ''').format(icon, ipnsKey)


class RequestInterceptor(QWebEngineUrlRequestInterceptor):
    def interceptRequest(self, info):
        pass


class CurrentPageHandler(BaseHandler):
    def __init__(self, parent):
        super().__init__(parent)

        self.rootCid = None

    @pyqtSlot()
    def getRootCid(self):
        return self.rootCid


class CustomWebPage (QtWebEngineWidgets.QWebEnginePage):
    jsConsoleMessage = pyqtSignal(int, str, int, str)

    def __init__(self, webProfile, parent):
        super(CustomWebPage, self).__init__(webProfile, parent)

    def registerPageHandler(self):
        self.channel = QWebChannel()
        self.setWebChannel(self.channel)
        self.pageHandler = CurrentPageHandler(self)
        self.channel.registerObject('gpage', self.pageHandler)

    def javaScriptConsoleMessage(self, level, message, lineNumber, sourceId):
        self.jsConsoleMessage.emit(level, message, lineNumber, sourceId)

    def registerProtocolHandlerRequestedDisabled(self, request):
        """
        registerProtocolHandlerRequested(request) can be used to
        handle JS 'navigator.registerProtocolHandler()' calls. If
        we want to enable that in the future it'll be done here.
        """

        scheme = request.scheme()
        origin = request.origin()

        qRet = questionBox(
            'registerProtocolHandler',
            'Allow {origin} to register procotol handler for: {h} ?'.format(
                origin=origin.toString(),
                h=scheme
            )
        )

        if qRet is True:
            request.accept()
        else:
            request.reject()

    async def render(self, tmpl, url=None, **ctx):
        if url:
            self.setHtml(await renderTemplate(tmpl, **ctx), url)
        else:
            self.setHtml(await renderTemplate(tmpl, **ctx))


class JSConsoleWidget(QTextBrowser):
    def __init__(self, parent):
        super(JSConsoleWidget, self).__init__(parent)

        self.setReadOnly(True)
        self.setAcceptRichText(True)
        self.setObjectName('javascriptConsole')
        self.append('<p><b>Javascript output console</b></p>')

    def showMessage(self, level, message, lineNumber, sourceId):
        sourcePath = IPFSPath(sourceId)

        if sourcePath.valid:
            source = str(sourcePath)
        else:
            source = sourceId

        if level == QWebEnginePage.InfoMessageLevel:
            levelH = 'INFO'
        elif level == QWebEnginePage.WarningMessageLevel:
            levelH = 'WARNING'
        elif level == QWebEnginePage.ErrorMessageLevel:
            levelH = 'ERROR'
        else:
            levelH = 'UNKNOWN'

        self.append('''
           <p>[{level}]
           <b>{source}</b> at line {line}: {message}
           </p>'''.format(
            source=source,
            level=levelH,
            line=lineNumber,
            message=message
        ))


class WebView(IPFSWebView):
    def __init__(self, browserTab, webProfile, enablePlugins=False,
                 parent=None):
        super(WebView, self).__init__(parent=parent)

        self.app = QCoreApplication.instance()
        self.browserTab = browserTab
        self.linkInfoTimer = QTimer()
        self.linkInfoTimer.timeout.connect(self.onLinkInfoTimeout)

        self.webPage = None
        self.changeWebProfile(webProfile)

        actionVSource = self.pageAction(QWebEnginePage.ViewSource)
        actionVSource.triggered.connect(self.onViewSource)

        self.webSettings = self.settings()
        self.webSettings.setAttribute(QWebEngineSettings.PluginsEnabled,
                                      enablePlugins)
        self.webSettings.setAttribute(QWebEngineSettings.LocalStorageEnabled,
                                      True)
        self.webSettings.setAttribute(
            QWebEngineSettings.LocalContentCanAccessFileUrls, True)
        self.webSettings.setAttribute(
            QWebEngineSettings.LocalContentCanAccessRemoteUrls, True)

    def createWindow(self, wintype):
        log.debug('createWindow called, wintype: {}'.format(wintype))

        # Disabled for now
        if wintype == QWebEnginePage.WebBrowserTab and 0:
            tab = self.app.mainWindow.addBrowserTab(current=False)
            return tab.webEngineView

    def onViewSource(self):
        def callback(html):
            if not html:
                return

            if self.browserTab.currentIpfsObject:
                tooltip = str(self.browserTab.currentIpfsObject)
            else:
                tooltip = self.page().url().toString()

            tab = GalacteekTab(self.app.mainWindow)
            widget = PageSourceWidget(tab)
            ensure(widget.showSource(html))
            tab.addToLayout(widget)
            self.app.mainWindow.registerTab(
                tab, 'Page source',
                current=True,
                icon=getMimeIcon('text/html'),
                tooltip=tooltip)

        self.page().toHtml(callback)

    def onJsMessage(self, level, message, lineNo, sourceId):
        self.browserTab.jsConsole.showMessage(level, message, lineNo, sourceId)

    def onLinkInfoTimeout(self):
        self.browserTab.ui.linkInfosLabel.setText('')

    def onFollowTheAtom(self, path):
        log.debug('Following atom feed: {}'.format(path))
        ensure(self.app.mainWindow.atomButton.atomFeedSubscribe(str(path)))

    def onLinkHovered(self, urlString):
        if not isinstance(urlString, str):
            return

        url = QUrl(urlString)
        if not url.isValid():
            return

        if url.isRelative() and self.browserTab.currentIpfsObject:
            ipfsPath = self.browserTab.currentIpfsObject.child(
                url.toString())
        else:
            ipfsPath = IPFSPath(urlString)

        if ipfsPath.valid:
            self.browserTab.ui.linkInfosLabel.setText(ipfsPath.fullPath)
        else:
            self.browserTab.ui.linkInfosLabel.setText(url.toString())

        if self.linkInfoTimer.isActive():
            self.linkInfoTimer.stop()

        self.linkInfoTimer.start(2200)

    def contextMenuCreateDefault(self):
        """
        Default context menu (returned for non-link rclicks)
        """
        menu = QMenu(self)
        menu.addAction(self.pageAction(QWebEnginePage.Back))
        menu.addAction(self.pageAction(QWebEnginePage.Forward))
        menu.addSeparator()
        menu.addAction(self.pageAction(QWebEnginePage.Reload))
        menu.addAction(self.pageAction(QWebEnginePage.ReloadAndBypassCache))
        menu.addSeparator()
        menu.addAction(self.pageAction(QWebEnginePage.Cut))
        menu.addAction(self.pageAction(QWebEnginePage.Copy))
        menu.addSeparator()
        menu.addAction(self.pageAction(QWebEnginePage.ViewSource))
        return menu

    def contextMenuEvent(self, event):
        # TODO: cleanup and refactoring

        analyzer = self.app.rscAnalyzer
        currentPage = self.page()
        contextMenuData = currentPage.contextMenuData()
        url = contextMenuData.linkUrl()
        mediaType = contextMenuData.mediaType()
        mediaUrl = contextMenuData.mediaUrl()

        if not url.isValid() and not mediaUrl.isValid():

            selectedText = self.webPage.selectedText()

            if selectedText:
                menu = QMenu(self)
                menu.addAction(iSaveSelectedText(), self.onSaveSelection)
                menu.exec(event.globalPos())
                return

            menu = self.contextMenuCreateDefault()
            menu.exec(event.globalPos())
            return

        ipfsPath = IPFSPath(url.toString(), autoCidConv=True)

        if mediaType != QWebEngineContextMenuData.MediaTypeNone and mediaUrl:
            mediaIpfsPath = IPFSPath(mediaUrl.toString(), autoCidConv=True)
        else:
            mediaIpfsPath = None

        if ipfsPath.valid:
            menu = QMenu(self)
            menu.addAction(getIcon('clipboard.png'),
                           iCopyPathToClipboard(),
                           functools.partial(self.app.setClipboardText,
                                             str(ipfsPath)
                                             ))
            menu.addSeparator()
            menu.addAction(getIcon('open.png'), iOpen(), functools.partial(
                ensure, self.app.resourceOpener.open(ipfsPath)))
            menu.addAction(getIcon('ipfs-logo-128-black.png'),
                           iOpenInTab(),
                           functools.partial(self.openInTab, ipfsPath))
            menu.addSeparator()
            menu.addAction(getIcon('hashmarks.png'),
                           iHashmark(),
                           functools.partial(self.hashmarkPath, str(ipfsPath)))
            menu.addSeparator()
            menu.addAction(getIcon('pin.png'), iPin(),
                           functools.partial(self.browserTab.pinPath,
                                             ipfsPath.objPath))
            menu.addAction(
                getIcon('pin.png'),
                iPinRecursive(),
                functools.partial(
                    self.browserTab.pinPath,
                    ipfsPath.objPath,
                    True))

            def rscAnalyzed(fut, path, iMenu):
                try:
                    mimeType, stat = fut.result()
                except Exception:
                    pass
                else:
                    if mimeType and mimeType.isAtomFeed:
                        iMenu.addSeparator()
                        iMenu.addAction(
                            getIcon('atom-feed.png'),
                            'Follow Atom feed',
                            functools.partial(self.onFollowTheAtom, path)
                        )
            ensure(
                analyzer(ipfsPath),
                futcallback=lambda fut: rscAnalyzed(
                    fut,
                    ipfsPath,
                    menu
                )
            )

            menu.exec(event.globalPos())

        elif mediaIpfsPath and mediaIpfsPath.valid:
            # Needs refactor
            menu = QMenu(self)

            menu.addSeparator()
            menu.addAction(getIcon('clipboard.png'),
                           iCopyPathToClipboard(),
                           functools.partial(self.app.setClipboardText,
                                             str(mediaIpfsPath)
                                             ))
            menu.addSeparator()
            menu.addAction(getIcon('open.png'), iOpen(), functools.partial(
                ensure, self.app.resourceOpener.open(str(mediaIpfsPath))))
            menu.addAction(
                getIcon('ipfs-logo-128-black.png'),
                iOpenInTab(),
                functools.partial(
                    self.openInTab,
                    mediaIpfsPath))
            menu.addSeparator()
            menu.addAction(getIcon('hashmarks.png'),
                           iHashmark(),
                           functools.partial(self.hashmarkPath,
                                             str(mediaIpfsPath)))
            menu.addSeparator()
            menu.addAction(getIcon('pin.png'), iPin(),
                           functools.partial(self.browserTab.pinPath,
                                             mediaIpfsPath.objPath))
            menu.addAction(getIcon('pin.png'), iPinRecursive(),
                           functools.partial(self.browserTab.pinPath,
                                             mediaIpfsPath.objPath, True))
            menu.addSeparator()

            async def rscMenuEnable(mainMenu, path, mimeType, stat, analyzer):
                if mimeType.isImage:
                    statInfo = StatInfo(stat)

                    if statInfo.valid and statInfo.dataLargerThan(
                            megabytes(4)):
                        return

                    codes = await analyzer.decodeQrCodes(str(path))
                    if codes:
                        codesMenu = qrCodesMenuBuilder(
                            codes, self.app.resourceOpener, parent=mainMenu)
                        mainMenu.addMenu(codesMenu)

            def rscAnalyzed(fut, path, mediaType, analyzer):
                try:
                    mimeType, stat = fut.result()
                except Exception:
                    pass
                else:
                    ensure(rscMenuEnable(menu, path, mimeType, stat,
                                         analyzer))

            ensure(
                analyzer(mediaIpfsPath),
                futcallback=lambda fut: rscAnalyzed(
                    fut,
                    mediaIpfsPath,
                    mediaType,
                    analyzer
                )
            )

            menu.exec(event.globalPos())
        else:
            # Non-IPFS URL

            menu = QMenu(self)
            scheme = url.scheme()

            if scheme in ['http', 'https'] and \
                    self.app.settingsMgr.allowHttpBrowsing:
                menu.addAction(
                    iOpenHttpInTab(),
                    functools.partial(self.openUrlInTab, url)
                )
            elif isSchemeRegistered(scheme):
                # Non-IPFS URL but a scheme we know
                menu.addAction(
                    iOpenLinkInTab(),
                    functools.partial(self.openUrlInTab, url)
                )

            menu.exec(event.globalPos())

    @asyncify
    async def onSaveSelection(self):
        selectedText = self.webPage.selectedText()

        if selectedText:
            await self.saveSelectedText(selectedText)

    @ipfsOp
    async def saveSelectedText(self, ipfsop, rawText):
        try:
            entry = await ipfsop.addString(rawText)
        except Exception:
            pass
        else:
            self.app.setClipboardText(entry['Hash'])

    def hashmarkPath(self, path):
        ipfsPath = IPFSPath(path, autoCidConv=True)
        if ipfsPath.valid:
            addHashmark(self.browserTab.app.marksLocal,
                        ipfsPath.fullPath,
                        ipfsPath.basename if ipfsPath.basename else iUnknown())

    def openInTab(self, path):
        tab = self.browserTab.gWindow.addBrowserTab()
        tab.browseFsPath(path)

    def openUrlInTab(self, url):
        tab = self.browserTab.gWindow.addBrowserTab()
        tab.enterUrl(url)

    def downloadLink(self, menudata):
        url = menudata.linkUrl()
        self.page().download(url, None)

    def installPage(self):
        if self.webPage:
            del self.webPage

        self.webPage = CustomWebPage(self.webProfile, self)
        self.webPage.jsConsoleMessage.connect(self.onJsMessage)
        self.webPage.linkHovered.connect(self.onLinkHovered)
        self.setPage(self.webPage)

    def changeWebProfile(self, profile):
        self.webProfile = profile
        self.installPage()


class BrowserKeyFilter(QObject):
    hashmarkPressed = pyqtSignal()
    savePagePressed = pyqtSignal()
    reloadPressed = pyqtSignal()
    reloadFullPressed = pyqtSignal()
    zoominPressed = pyqtSignal()
    zoomoutPressed = pyqtSignal()
    focusUrlPressed = pyqtSignal()
    findInPagePressed = pyqtSignal()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress:
            modifiers = event.modifiers()
            key = event.key()

            if modifiers & Qt.ControlModifier:
                if key == Qt.Key_B:
                    self.hashmarkPressed.emit()
                    return True
                if key == Qt.Key_R:
                    self.reloadPressed.emit()
                    return True
                if key == Qt.Key_F5:
                    self.reloadFullPressed.emit()
                    return True
                if key == Qt.Key_L:
                    self.focusUrlPressed.emit()
                    return True
                if key == Qt.Key_Plus:
                    self.zoominPressed.emit()
                    return True
                if key == Qt.Key_Minus:
                    self.zoomoutPressed.emit()
                    return True
                if key == Qt.Key_F:
                    self.findInPagePressed.emit()
                    return True

            if key == Qt.Key_F5:
                self.reloadPressed.emit()
                return True

        return False


class URLInputWidget(QLineEdit):
    def __init__(self, history, historyView, parent):
        super(URLInputWidget, self).__init__(parent)

        self.app = QCoreApplication.instance()
        self.history = history
        self.historyMatches = historyView
        self.browser = parent

        self.setObjectName('urlZone')
        self.setMinimumWidth(400)
        self.setDragEnabled(True)
        self.setMaxLength(1024)

        self.urlEditing = False
        self.urlInput = None

        self.editTimer = QTimer(self)
        self.editTimer.timeout.connect(self.onTimeoutUrlEdit)
        self.editTimer.setSingleShot(True)

        self.returnPressed.connect(self.onReturnPressed)
        self.historyMatches.historyItemSelected.connect(
            self.onHistoryItemSelected)
        self.historyMatches.collapsed.connect(
            self.onHistoryCollapse)
        self.textEdited.connect(self.onUrlUserEdit)

    @property
    def editTimeoutMs(self):
        timeout = self.app.settingsMgr.urlHistoryEditTimeout
        return timeout if isinstance(timeout, int) else 1000

    def unfocus(self):
        self.urlEditing = False
        self.clearFocus()

    def cancelTimer(self):
        self.editTimer.stop()

    def hideMatches(self):
        self.historyMatches.hide()

    def onReturnPressed(self):
        self.urlEditing = False
        self.unfocus()
        self.browser.handleEditedUrl(self.text())

    def onHistoryCollapse(self):
        self.setFocus(Qt.PopupFocusReason)
        self.deselect()

    def focusInEvent(self, event):
        if event.reason() in [Qt.ShortcutFocusReason, Qt.MouseFocusReason,
                              Qt.PopupFocusReason, Qt.ActiveWindowFocusReason,
                              Qt.TabFocusReason]:
            self.urlEditing = True

        super(URLInputWidget, self).focusInEvent(event)

    def focusOutEvent(self, event):
        if event.reason() not in [
                Qt.ActiveWindowFocusReason,
                Qt.PopupFocusReason,
                Qt.TabFocusReason,
                Qt.OtherFocusReason]:
            self.urlEditing = False
            self.editTimer.stop()

        super(URLInputWidget, self).focusOutEvent(event)

    def onUrlUserEdit(self, text):
        self.urlInput = text

        if not self.urlEditing:
            return

        if not self.editTimer.isActive():
            self.editTimer.start(self.editTimeoutMs)
        else:
            self.editTimer.stop()
            self.editTimer.start(self.editTimeoutMs)

    def onTimeoutUrlEdit(self):
        ensure(self.historyLookup())

    async def historyLookup(self):
        if self.urlInput:
            markMatches = []

            async for mark in self.app.marksLocal.searchAllByMetadata({
                'title': self.urlInput,
                'description': self.urlInput
            }):
                try:
                    markMatches.append({
                        'title': mark.markData['metadata']['title'],
                        'url': IPFSPath(mark.path).ipfsUrl
                    })
                except Exception:
                    continue

            hMatches = await self.history.match(self.urlInput)

            if len(markMatches) > 0 or len(hMatches) > 0:
                self.historyMatches.showMatches(markMatches, hMatches)
                self.historyMatches.show()
                self.historyMatches.setFocus(Qt.OtherFocusReason)

            self.editTimer.stop()

    def onHistoryItemSelected(self, urlStr):
        self.historyMatches.hide()
        self.urlEditing = False

        url = QUrl(urlStr)

        if url.isValid():
            self.browser.enterUrl(url)


class CurrentObjectController(PopupToolButton):
    objectVisited = pyqtSignal(IPFSPath)

    dagViewRequested = pyqtSignal(IPFSPath)

    def __init__(self, parent=None):
        super(CurrentObjectController, self).__init__(
            mode=QToolButton.MenuButtonPopup, parent=parent)

        self.app = QApplication.instance()

        self.currentPath = None
        self.objectVisited.connect(self.onVisited)

        self.pathCubeOrange = ':/share/icons/cube-orange.png'
        self.pathCubeBlue = ':/share/icons/cube-blue.png'

        self.setIcon(getIcon('cube-orange.png'))
        self.setAutoRaise(True)

        self.dagViewAction = QAction(getIcon('ipld.png'),
                                     iDagView(), self,
                                     triggered=self.onDAGView)
        self.parentDagViewAction = QAction(getIcon('ipld.png'),
                                           iParentDagView(), self,
                                           triggered=self.onParentDAGView)
        self.qaLinkAction = QAction(getIconIpfsWhite(),
                                    iLinkToQaToolbar(), self,
                                    triggered=self.onQaLink)
        self.copyCurPathAction = QAction(getIcon('clipboard.png'),
                                         iCopyPathToClipboard(), self,
                                         triggered=self.onCopyPathToCb)
        self.hashmarkRootCidAction = QAction(getIcon('hashmark-black.png'),
                                             iHashmarkRootObject(), self,
                                             triggered=self.onHashmarkRoot)
        self.hashmarkObjectAction = QAction(getIcon('hashmark-black.png'),
                                            iHashmarkThisPage(), self,
                                            triggered=self.onHashmarkThisPage)

        self.setDefaultAction(self.dagViewAction)

        self.menu.addAction(self.copyCurPathAction)
        self.menu.addSeparator()
        self.menu.addAction(self.qaLinkAction)
        self.menu.addSeparator()
        self.menu.addAction(self.dagViewAction)
        self.menu.addSeparator()
        self.menu.addAction(self.parentDagViewAction)
        self.menu.addSeparator()
        self.menu.addAction(self.hashmarkRootCidAction)
        self.menu.addSeparator()
        self.menu.addAction(self.hashmarkObjectAction)
        self.menu.addSeparator()

    def onQaLink(self):
        if self.currentPath:
            self.app.mainWindow.qaToolbar.ipfsObjectDropped.emit(
                self.currentPath)

    def onDAGView(self):
        self.dagViewRequested.emit(self.currentPath)

    def onParentDAGView(self):
        self.dagViewRequested.emit(self.currentPath.parent())

    def onCopyPathToCb(self):
        if self.currentPath:
            self.app.setClipboardText(str(self.currentPath))

    def onHashmarkThisPage(self):
        if self.currentPath:
            addHashmark(self.app.marksLocal,
                        self.currentPath.fullPath,
                        '')

    def onHashmarkRoot(self):
        if self.currentPath:
            if self.currentPath.isIpfs:
                rootCidPath = joinIpfs(self.currentPath.rootCidRepr)
                addHashmark(self.app.marksLocal,
                            rootCidPath,
                            rootCidPath)
            elif self.currentPath.isIpns:
                root = self.currentPath.root()
                if not root:
                    return messageBox(
                        iInvalidObjectPath(str(self.currentPath)))

                addHashmark(self.app.marksLocal,
                            str(root), str(root))

    def onVisited(self, ipfsPath):
        ensure(self.displayObjectInfos(ipfsPath))

    @ipfsOp
    async def displayObjectInfos(self, ipfsop, ipfsPath):
        self.currentPath = ipfsPath

        self.parentDagViewAction.setEnabled(
            self.currentPath.subPath is not None)

        objCid = None
        cidRepr = ipfsPath.rootCidRepr

        if ipfsPath.isIpfs:
            # First generic tooltip without resolved CID
            self.setToolTip(iCidTooltipMessage(
                self.pathCubeBlue,
                ipfsPath.rootCid.version, '',
                cidRepr,
                iUnknown()
            ))

            objCid = await ipfsPath.resolve(ipfsop, timeout=5)

        if self.currentPath != ipfsPath:
            # Current object has changed while resolving, let the new
            # object's tooltip appear
            return

        if ipfsPath.rootCidUseB32:
            self.setIcon(getIcon('cube-orange.png'))

            self.setToolTip(iCidTooltipMessage(
                self.pathCubeOrange,
                1, '(base32)',
                cidRepr,
                objCid if objCid else iUnknown()
            ))
        else:
            self.setIcon(getIcon('cube-blue.png'))

            if ipfsPath.rootCid:
                self.setToolTip(iCidTooltipMessage(
                    self.pathCubeBlue,
                    ipfsPath.rootCid.version, '',
                    cidRepr,
                    objCid if objCid else iUnknown()
                ))

            elif ipfsPath.isIpns:
                self.setToolTip(
                    iIpnsTooltipMessage(
                        self.pathCubeBlue,
                        ipfsPath.ipnsFqdn if ipfsPath.ipnsFqdn else
                        ipfsPath.ipnsKey
                    )
                )


class BrowserTab(GalacteekTab):
    # signals
    ipfsObjectVisited = pyqtSignal(IPFSPath)

    def __init__(self, gWindow, pinBrowsed=False):
        super(BrowserTab, self).__init__(gWindow)

        self.browserWidget = QWidget(self)
        self.browserWidget.setAttribute(Qt.WA_DeleteOnClose)

        self.jsConsoleWidget = JSConsoleWidget(self)
        self.jsConsoleWidget.setMaximumHeight(180)
        self.jsConsoleWidget.hide()

        self.ui = ui_browsertab.Ui_BrowserTabForm()
        self.ui.setupUi(self.browserWidget)

        self.vLayout.addWidget(self.browserWidget)
        self.ui.vLayoutConsole.addWidget(self.jsConsoleWidget)

        # URL zone
        self.historySearches = HistoryMatchesWidget(parent=self)
        self.historySearches.setMinimumHeight(160)
        self.historySearches.hide()
        self.urlZone = URLInputWidget(self.history, self.historySearches, self)
        self.ui.layoutHistory.addWidget(self.historySearches)
        self.ui.layoutUrl.addWidget(self.urlZone)

        # Search
        self.ui.searchInPage.returnPressed.connect(self.onSearchInPage)
        self.ui.searchInPage.textEdited.connect(self.onSearchInPageEdit)
        self.ui.searchPrev.clicked.connect(self.onSearchInPagePrev)
        self.ui.searchNext.clicked.connect(self.onSearchInPageNext)
        self.ui.searchButton.clicked.connect(self.onSearchInPage)
        self.ui.searchCancel.clicked.connect(self.onSearchCancel)
        self.ui.findToolButton.setCheckable(True)
        self.ui.findToolButton.toggled.connect(self.searchControlsSetVisible)
        self.searchControlsSetVisible(False)

        # Current object controller
        self.curObjectCtrl = CurrentObjectController(self)
        self.curObjectCtrl.dagViewRequested.connect(self.onDagViewRequested)
        self.curObjectCtrl.hide()
        self.ui.layoutObjectInfo.addWidget(self.curObjectCtrl)

        initialProfileName = self.app.settingsMgr.defaultWebProfile
        if initialProfileName not in self.app.availableWebProfilesNames():
            initialProfileName = WP_NAME_MINIMAL

        webProfile = self.getWebProfileByName(initialProfileName)

        self.webEngineView = WebView(
            self, webProfile,
            enablePlugins=self.app.settingsMgr.ppApiPlugins,
            parent=self
        )
        self.ui.vLayoutBrowser.addWidget(self.webEngineView)

        self.webEngineView.urlChanged.connect(self.onUrlChanged)
        self.webEngineView.loadFinished.connect(self.onLoadFinished)
        self.webEngineView.iconChanged.connect(self.onIconChanged)
        self.webEngineView.loadProgress.connect(self.onLoadProgress)
        self.webEngineView.titleChanged.connect(self.onTitleChanged)

        self.ui.backButton.clicked.connect(self.backButtonClicked)
        self.ui.forwardButton.clicked.connect(self.forwardButtonClicked)

        self.ui.reloadPageButton.clicked.connect(self.refreshButtonClicked)
        self.ui.stopButton.clicked.connect(self.stopButtonClicked)

        self.ui.backButton.setEnabled(False)
        self.ui.forwardButton.setEnabled(False)
        self.ui.stopButton.setEnabled(False)

        self.ui.linkInfosLabel.setObjectName('linkInfos')

        # Page ops button
        self.createPageOpsButton()

        self.hashmarkPageAction = QAction(getIcon('hashmark-black.png'),
                                          iHashmarkThisPage(), self,
                                          shortcut=QKeySequence('Ctrl+b'),
                                          triggered=self.onHashmarkPage)
        self.ui.hashmarkThisPage.setDefaultAction(self.hashmarkPageAction)
        self.ui.hashmarkThisPage.setEnabled(False)
        self.ui.hashmarkThisPage.setAutoRaise(True)

        self.ui.hLayoutCtrl.addWidget(self.pageOpsButton)

        # Setup the IPFS control tool button (has actions
        # for browsing CIDS and web profiles etc..)

        self.ipfsControlMenu = QMenu(self)
        self.webProfilesMenu = QMenu('Web profile')

        self.webProfilesGroup = QActionGroup(self.webProfilesMenu)
        self.webProfilesGroup.setExclusive(True)
        self.webProfilesGroup.triggered.connect(self.onWebProfileSelected)

        self.webProMinAction = QAction(WP_NAME_MINIMAL, self)
        self.webProMinAction.setCheckable(True)
        self.webProIpfsAction = QAction(WP_NAME_IPFS, self)
        self.webProIpfsAction.setCheckable(True)
        self.webProWeb3Action = QAction(WP_NAME_WEB3, self)
        self.webProWeb3Action.setCheckable(True)

        self.webProfilesGroup.addAction(self.webProMinAction)
        self.webProfilesGroup.addAction(self.webProIpfsAction)
        self.webProfilesGroup.addAction(self.webProWeb3Action)

        for action in self.webProfilesGroup.actions():
            self.webProfilesMenu.addAction(action)

        self.checkWebProfileByName(initialProfileName)

        self.browserHelpAction = QAction(getIcon('help.png'),
                                         iHelp(), self,
                                         triggered=self.onShowBrowserHelp)
        self.loadIpfsCIDAction = QAction(getIconIpfsIce(),
                                         iBrowseIpfsCID(), self,
                                         triggered=self.onLoadIpfsCID)
        self.loadIpfsMultipleCIDAction = QAction(
            getIconIpfsIce(),
            iBrowseIpfsMultipleCID(),
            self,
            triggered=self.onLoadIpfsMultipleCID)
        self.loadIpnsAction = QAction(getIconIpfsWhite(),
                                      iBrowseIpnsHash(), self,
                                      triggered=self.onLoadIpns)
        self.followIpnsAction = QAction(getIconIpfsWhite(),
                                        iFollow(), self,
                                        triggered=self.onFollowIpns)
        self.loadHomeAction = QAction(getIcon('go-home.png'),
                                      iBrowseHomePage(), self,
                                      triggered=self.onLoadHome)

        self.jsConsoleAction = QAction(getIcon('terminal.png'),
                                       iJsConsole(), self,
                                       shortcut=QKeySequence('Ctrl+x'))
        self.jsConsoleAction.setCheckable(True)
        self.jsConsoleAction.toggled.connect(self.onJsConsoleToggle)

        self.mapPathAction = QAction(getIconIpfsIce(),
                                     iCreateQaMapping(), self,
                                     triggered=self.onCreateMapping)

        self.loadFromCbAction = QAction(
            getIcon('clipboard.png'),
            iBrowseCurrentClipItem(), self,
            triggered=self.onBrowseCurrentClipItem
        )

        self.followIpnsAction.setEnabled(False)

        self.ipfsControlMenu.addAction(self.browserHelpAction)
        self.ipfsControlMenu.addSeparator()
        self.ipfsControlMenu.addMenu(self.webProfilesMenu)
        self.ipfsControlMenu.addSeparator()

        self.ipfsControlMenu.addAction(self.jsConsoleAction)
        self.ipfsControlMenu.addSeparator()
        self.ipfsControlMenu.addAction(self.mapPathAction)
        self.ipfsControlMenu.addSeparator()
        self.ipfsControlMenu.addAction(self.loadFromCbAction)
        self.ipfsControlMenu.addAction(self.loadIpfsCIDAction)
        self.ipfsControlMenu.addAction(self.loadIpfsMultipleCIDAction)
        self.ipfsControlMenu.addAction(self.loadIpnsAction)
        self.ipfsControlMenu.addAction(self.followIpnsAction)
        self.ipfsControlMenu.addAction(self.loadHomeAction)

        self.ui.loadIpfsButton.setMenu(self.ipfsControlMenu)
        self.ui.loadIpfsButton.setPopupMode(QToolButton.InstantPopup)
        self.ui.loadIpfsButton.clicked.connect(self.onLoadIpfsCID)

        self.ui.pBarBrowser.setTextVisible(False)
        self.ui.pinAllButton.setCheckable(True)
        self.ui.pinAllButton.setAutoRaise(True)

        if pinBrowsed:
            self.ui.pinAllButton.setChecked(True)
        else:
            self.ui.pinAllButton.setChecked(self.gWindow.pinAllGlobalChecked)

        self.ui.pinAllButton.toggled.connect(self.onToggledPinAll)

        # PIN tool button
        iconPin = getIcon('pin.png')

        pinMenu = QMenu(self)
        pinMenu.addAction(iconPin, iPinThisPage(), self.onPinSingle)
        pinMenu.addAction(iconPin, iPinRecursive(), self.onPinRecursive)
        pinMenu.addSeparator()
        pinMenu.addAction(iconPin, iPinRecursiveParent(),
                          self.onPinRecursiveParent)
        pinMenu.addSeparator()
        pinMenu.addAction(iconPin, iPinPageLinks(), self.onPinPageLinks)
        pinMenu.addSeparator()
        pinMenu.addAction(getIcon('help.png'), iHelp(), functools.partial(
            self.app.manuals.browseManualPage, 'pinning.html'))

        self.ui.pinToolButton.setMenu(pinMenu)
        self.ui.pinToolButton.setIcon(iconPin)
        self.ui.pinToolButton.setText(iPin())

        self.ui.zoomInButton.clicked.connect(self.onZoomIn)
        self.ui.zoomOutButton.clicked.connect(self.onZoomOut)

        # Event filter
        evfilter = BrowserKeyFilter(self)
        evfilter.hashmarkPressed.connect(self.onHashmarkPage)
        evfilter.reloadPressed.connect(self.onReloadPage)
        evfilter.reloadFullPressed.connect(self.onReloadPageNoCache)
        evfilter.zoominPressed.connect(self.onZoomIn)
        evfilter.zoomoutPressed.connect(self.onZoomOut)
        evfilter.focusUrlPressed.connect(self.onFocusUrl)
        evfilter.findInPagePressed.connect(self.onToggleSearchControls)
        self.installEventFilter(evfilter)

        self.resolveTimer = QTimer(self)
        self.resolveTimerEnsMs = 7000
        self.urlInput = None

        self.app.clipTracker.clipboardPathProcessed.connect(
            self.onClipboardIpfs)
        self.ipfsObjectVisited.connect(self.onPathVisited)

        self._currentIpfsObject = None
        self._currentUrl = None
        self.currentPageTitle = None

    @property
    def history(self):
        return self.gWindow.app.urlHistory

    @property
    def webView(self):
        return self.webEngineView

    @property
    def currentUrl(self):
        return self._currentUrl

    @property
    def currentUrlHasFileName(self):
        if self.currentUrl:
            return self.currentUrl.fileName() != ''

    @property
    def jsConsole(self):
        return self.jsConsoleWidget

    @property
    def tabPage(self):
        return self.gWindow.tabWidget.widget(self.tabPageIdx)

    @property
    def tabPageIdx(self):
        return self.gWindow.tabWidget.indexOf(self)

    @property
    def gatewayAuthority(self):
        return self.app.gatewayAuthority

    @property
    def gatewayUrl(self):
        return self.app.gatewayUrl

    @property
    def pinAll(self):
        return self.ui.pinAllButton.isChecked()

    @property
    def currentIpfsObject(self):
        return self._currentIpfsObject

    def createPageOpsButton(self):
        icon = getMimeIcon('text/html')
        self.pageOpsButton = PopupToolButton(
            icon, mode=QToolButton.InstantPopup, parent=self
        )
        self.pageOpsButton.setAutoRaise(True)
        self.pageOpsButton.menu.addAction(
            icon, iSaveContainedWebPage(), self.onSavePageContained)

        ensure(self.createPageOpsMfsMenu())

    @ipfsOp
    async def createPageOpsMfsMenu(self, ipfsop):
        self.mfsMenu = ipfsop.ctx.currentProfile.createMfsMenu(
            title=iLinkToMfsFolder(), parent=self
        )
        self.mfsMenu.setIcon(getIcon('folder-open-black.png'))
        self.mfsMenu.triggered.connect(self.onMoveToMfsMenuTriggered)
        self.pageOpsButton.menu.addSeparator()
        self.pageOpsButton.menu.addMenu(self.mfsMenu)

    def onMoveToMfsMenuTriggered(self, action):
        mfsItem = action.data()

        if not mfsItem:
            return

        if not self.currentIpfsObject:
            # ERR
            return

        cTitle = self.currentPageTitle if self.currentPageTitle else ''

        runDialog(
            TitleInputDialog, cTitle,
            accepted=functools.partial(
                self.onMoveToMfsWithTitle,
                self.currentIpfsObject,
                mfsItem
            )
        )

    def onMoveToMfsWithTitle(self, ipfsPath, mfsItem, dialog):
        cTitle = dialog.tEdit.text().replace('/', '_')
        if not cTitle:
            return messageBox('Invalid title')

        ensure(self.linkPageToMfs(ipfsPath, mfsItem, cTitle))

    @ipfsOp
    async def linkPageToMfs(self, ipfsop, ipfsPath, mfsItem, cTitle):
        basename = ipfsPath.basename

        dest = os.path.join(mfsItem.path, cTitle.strip())

        if basename and not (ipfsPath.isIpfsRoot or ipfsPath.isIpnsRoot):
            await ipfsop.filesMkdir(dest)
            dest = os.path.join(dest, basename)

        try:
            await ipfsop.client.files.cp(
                ipfsPath.objPath,
                dest
            )
        except aioipfs.APIError:
            messageBox('Error while linking')

        self.pinPath(ipfsPath.objPath, recursive=True, notify=True)

    def installCustomPageScripts(self, handler, url):
        # Page-specific IPFS scripts

        if url.scheme() == SCHEME_IPFS:
            rootCid = None

            if issubclass(handler.__class__, DAGProxySchemeHandler):
                if handler.proxied:
                    rootCid = handler.proxied.dagCid
            elif issubclass(handler.__class__, MultiDAGProxySchemeHandler):
                if len(handler.proxied) > 0:
                    rootCid = handler.proxied[0].dagCid
            else:
                rootCid = url.host()

            if not rootCid:
                return

            scripts = self.webEngineView.webPage.scripts()

            script = scriptFromString('rootcid', '''
                window.rootcid = '{cid}';
            '''.format(cid=rootCid))

            existing = scripts.findScript('rootcid')
            if existing:
                scripts.remove(existing)

            scripts.insert(script)

    def onDagViewRequested(self, ipfsPath):
        view = DAGViewer(ipfsPath.objPath, self.app.mainWindow)
        self.app.mainWindow.registerTab(
            view, iDagViewer(),
            current=True,
            icon=getIcon('ipld.png'),
            tooltip=str(ipfsPath)
        )

    def onPathVisited(self, ipfsPath):
        # Called after a new IPFS object has been loaded in this tab

        if ipfsPath.valid:
            self.curObjectCtrl.objectVisited.emit(ipfsPath)
            ensure(self.tsVisitPath(ipfsPath.objPath))

    @ipfsStatOp
    async def tsVisitPath(self, ipfsop, path, stat):
        # If automatic pinning is checked we pin the object
        if self.pinAll is True:
            self.pinPath(path, recursive=False, notify=False)

    def changeWebProfileByName(self, wpName):
        if self.webEngineView.webProfile.profileName == wpName:
            log.debug('Profile {p} already set'.format(p=wpName))
            return

        self.checkWebProfileByName(wpName)

    def checkWebProfileByName(self, wpName):
        if wpName == WP_NAME_MINIMAL:
            self.webProMinAction.setChecked(True)
        elif wpName == WP_NAME_IPFS:
            self.webProIpfsAction.setChecked(True)
        elif wpName == WP_NAME_WEB3:
            self.webProWeb3Action.setChecked(True)

    def getWebProfileByName(self, wpName, fallback=WP_NAME_MINIMAL):
        return self.app.webProfiles.get(
            wpName, self.app.webProfiles[fallback])

    def onWebProfileSelected(self, action):
        if self.webEngineView.webProfile.profileName == action.text():
            log.debug('Profile {p} already set'.format(p=action.text()))
            return

        url = self.currentUrl

        if action is self.webProMinAction:
            self.webEngineView.changeWebProfile(
                self.getWebProfileByName(WP_NAME_MINIMAL)
            )
        elif action is self.webProIpfsAction:
            self.webEngineView.changeWebProfile(
                self.getWebProfileByName(WP_NAME_IPFS)
            )
        elif action is self.webProWeb3Action:
            self.webEngineView.changeWebProfile(
                self.getWebProfileByName(WP_NAME_WEB3)
            )

        if url:
            # Reload page
            self.enterUrl(url)
        else:
            async def showProfileChangedPage():
                pName = self.webEngineView.webProfile.profileName
                body = await renderTemplate(
                    'profilechanged.html',
                    webProfile=pName,
                    title='Web profile changed to {name}'.format(name=pName)
                )
                self.webEngineView.webPage.setHtml(
                    body, QUrl('z:/profilechanged.html'))

            ensure(showProfileChangedPage())

    def onShowBrowserHelp(self):
        self.app.manuals.browseManualPage('browsing.html')

    def searchControlsSetVisible(self, view):
        for w in [self.ui.searchInPage,
                  self.ui.searchCancel,
                  self.ui.searchButton,
                  self.ui.searchLabel,
                  self.ui.searchPrev,
                  self.ui.searchNext]:
            w.setVisible(view)

        if view is True:
            self.ui.searchInPage.setFocus(Qt.OtherFocusReason)

        self.ui.findToolButton.setChecked(view)

    def onToggleSearchControls(self):
        self.searchControlsSetVisible(not self.ui.searchInPage.isVisible())

    def onSearchInPageEdit(self, text):
        self.ui.searchInPage.setStyleSheet(
            'QLineEdit { color: black; }'
        )

    def onSearchInPage(self):
        self.searchInPage()

    def searchReset(self):
        self.ui.searchInPage.clear()
        self.searchInPage()

    def searchInPage(self, flags=0):
        text = self.ui.searchInPage.text()
        self.webEngineView.webPage.findText(
            text, QWebEnginePage.FindFlags(flags),
            self.searchCallback
        )

    def searchCallback(self, found):
        if not found:
            self.ui.searchInPage.setStyleSheet(
                'QLineEdit { color: red; }'
            )
        else:
            self.ui.searchInPage.setStyleSheet(
                'QLineEdit { color: black; }'
            )

    def onSearchInPageNext(self):
        self.searchInPage()

    def onSearchInPagePrev(self):
        self.searchInPage(QWebEnginePage.FindBackward)

    def onSearchCancel(self):
        self.searchReset()
        self.searchControlsSetVisible(False)

    def onJsConsoleToggle(self, checked):
        self.jsConsoleWidget.setVisible(checked)

    def onCreateMapping(self):
        if not self.currentIpfsObject:
            return messageBox(iNotAnIpfsResource())

        runDialog(QSchemeCreateMappingDialog, self.currentIpfsObject)

    def onClipboardIpfs(self, pathObj):
        pass

    def onToggledPinAll(self, checked):
        pass

    def onFocusUrl(self):
        self.focusUrlZone(True, reason=Qt.ShortcutFocusReason)

    def focusUrlZone(self, select=False, reason=Qt.OtherFocusReason):
        self.urlZone.setFocus(reason)

        if select:
            self.urlZone.setSelection(0, len(self.urlZone.text()))

        self.urlZone.urlEditing = True

    def onReloadPage(self):
        self.reloadPage()

    def onReloadPageNoCache(self):
        self.reloadPage(bypassCache=True)

    def reloadPage(self, bypassCache=False):
        self.urlZone.unfocus()
        self.webEngineView.triggerPageAction(
            QWebEnginePage.ReloadAndBypassCache if bypassCache is True else
            QWebEnginePage.Reload
        )

    def onSavePageContained(self):
        tmpd = self.app.tempDirCreate(self.app.tempDirWeb)
        if tmpd is None:
            return messageBox('Cannot create temporary directory')

        page = self.webView.page()

        url = page.url()
        filename = url.fileName()

        if not filename or cidValid(filename):
            filename = 'index.html'

        path = os.path.join(tmpd, filename)
        page.save(path, QWebEngineDownloadItem.CompleteHtmlSaveFormat)

    def onHashmarkPage(self):
        if self.currentIpfsObject and self.currentIpfsObject.valid:
            addHashmark(self.app.marksLocal,
                        self.currentIpfsObject.fullPath,
                        self.currentPageTitle,
                        stats=self.app.ipfsCtx.objectStats.get(
                            self.currentIpfsObject.path, {}))
        else:
            messageBox(iNotAnIpfsResource())

    def onPinResult(self, f):
        try:
            path, code, msg = f.result()
        except:
            pass
        else:
            path = IPFSPath(path)
            if not path.valid:
                log.debug('Invalid path in pin result: {}'.format(str(path)))
                return

            if code == 0:
                self.app.systemTrayMessage('PIN', iPinSuccess(str(path)),
                                           timeout=2000)
            elif code == 1:
                self.app.systemTrayMessage('PIN', iPinError(str(path), msg),
                                           timeout=3000)
            elif code == 2:
                # Cancelled, no need to notify here
                pass
            else:
                log.debug('Unknown status code for pinning result')

    @ipfsOp
    async def pinQueuePath(self, ipfsop, path, recursive, notify):
        log.debug('Pinning object {0} (recursive: {1})'.format(path,
                                                               recursive))
        onSuccess = None
        if notify is True:
            onSuccess = self.onPinResult
        await ipfsop.ctx.pinner.queue(path, recursive, onSuccess,
                                      qname='browser')

    def pinPath(self, path, recursive=True, notify=True):
        if isinstance(path, str):
            ensure(self.pinQueuePath(path, recursive, notify))
        elif isinstance(path, IPFSPath):
            ensure(self.pinQueuePath(path.objPath, recursive, notify))

    def printButtonClicked(self):
        # TODO: reinstate the printing button
        printer = QPrinter()
        dialog = QPrintDialog(printer, self)
        dialog.setModal(True)

        def success(ok):
            return ok

        if dialog.exec_() == QDialog.Accepted:
            currentPage = self.webEngineView.page()
            currentPage.print(printer, success)

    def onPinSingle(self):
        if not self.currentIpfsObject:
            return messageBox(iNotAnIpfsResource())

        self.pinPath(self.currentIpfsObject.objPath, recursive=False)

    def onPinRecursive(self):
        if not self.currentIpfsObject:
            return messageBox(iNotAnIpfsResource())

        self.pinPath(self.currentIpfsObject.objPath, recursive=True)

    def onPinRecursiveParent(self):
        if not self.currentIpfsObject:
            return messageBox(iNotAnIpfsResource())

        if self.currentUrlHasFileName:
            parent = self.currentIpfsObject.parent()
            self.pinPath(parent.objPath, recursive=True)
        else:
            self.pinPath(self.currentIpfsObject.objPath, recursive=True)

    def onPinPageLinks(self):
        if not self.currentIpfsObject:
            return messageBox(iNotAnIpfsResource())

        def htmlReady(htmlCode):

            ensure(self.pinIpfsLinksInPage(htmlCode))

        self.webEngineView.webPage.toHtml(htmlReady)

    @ipfsOp
    async def pinIpfsLinksInPage(self, ipfsop, htmlCode):
        if not isinstance(htmlCode, str):
            return

        baseUrl = self.currentUrl.url(
            QUrl.RemoveFilename | QUrl.RemoveFragment | QUrl.RemoveQuery)
        basePath = IPFSPath(baseUrl)

        if not basePath.valid:
            return

        # Being a potentially CPU-intensive task for large documents, run
        # the HTML parser in the threadpool executor

        parser = IPFSLinksParser(basePath)
        logUser.info('Scanning links in object: {op}'.format(
            op=basePath))

        await self.app.loop.run_in_executor(self.app.executor,
                                            parser.feed, htmlCode)

        tab = PinBatchTab(self.app.mainWindow)
        bulkWidget = PinBatchWidget(basePath, parser.links, parent=tab)
        tab.addToLayout(bulkWidget)
        self.app.mainWindow.registerTab(tab, 'Pin batch', current=True,
                                        icon=getIcon('pin.png'))

    def onBrowseCurrentClipItem(self):
        current = self.app.clipTracker.getCurrent()
        if current:
            self.browseFsPath(IPFSPath(current.fullPath))
        else:
            messageBox(iClipboardEmpty())

    def onLoadIpfsCID(self):
        def onValidated(d):
            path = IPFSPath(d.getHash(), autoCidConv=True)
            if path.valid:
                self.browseFsPath(path)

        runDialog(IPFSCIDInputDialog, title=iEnterIpfsCIDDialog(),
                  accepted=onValidated)

    def onLoadIpfsMultipleCID(self):
        def onValidated(dlg):
            # Open a tab for every CID
            cids = dlg.getCIDs()
            for cid in cids:
                path = IPFSPath(cid, autoCidConv=True)
                if not path.valid:
                    continue
                self.gWindow.addBrowserTab().browseFsPath(path)

        runDialog(IPFSMultipleCIDInputDialog,
                  title=iEnterIpfsCIDDialog(),
                  accepted=onValidated)

    def onLoadIpns(self):
        text, ok = QInputDialog.getText(self,
                                        iEnterIpnsDialog(),
                                        iEnterIpns())
        if ok:
            self.browseIpnsKey(text)

    def onFollowIpns(self):
        if self.currentIpfsObject and self.currentIpfsObject.isIpns:
            root = self.currentIpfsObject.root().objPath
            runDialog(AddFeedDialog, self.app.marksLocal,
                      root,
                      title=iFollowIpnsDialog())

    def onLoadHome(self):
        self.loadHomePage()

    def loadHomePage(self):
        homeUrl = self.app.settingsMgr.getSetting(CFG_SECTION_BROWSER,
                                                  CFG_KEY_HOMEURL)
        self.enterUrl(QUrl(homeUrl))

    def refreshButtonClicked(self):
        self.webEngineView.reload()

    def stopButtonClicked(self):
        self.webEngineView.stop()
        self.ui.pBarBrowser.setValue(0)
        self.ui.stopButton.setEnabled(False)

    def backButtonClicked(self):
        self.urlZone.unfocus()
        currentPage = self.webEngineView.page()
        currentPage.history().back()

    def forwardButtonClicked(self):
        self.urlZone.unfocus()
        currentPage = self.webEngineView.page()
        currentPage.history().forward()

    # def onUrlChanged(self, url):
    @asyncify
    async def onUrlChanged(self, url):
        if not url.isValid() or url.scheme() in ['data', SCHEME_Z]:
            return

        sHandler = self.webEngineView.webProfile.urlSchemeHandler(
            url.scheme().encode())

        if url.scheme() in [SCHEME_IPFS, SCHEME_IPNS, SCHEME_DWEB]:
            # ipfs:// or ipns://
            self.urlZone.setStyleSheet('''
                QLineEdit {
                    background-color: #C3D7DF;
                }''')

            self.urlZone.clear()
            self.urlZone.insert(url.toString())
            self._currentIpfsObject = IPFSPath(url.toString())
            self._currentUrl = url
            self.ipfsObjectVisited.emit(self.currentIpfsObject)
            self.ui.hashmarkThisPage.setEnabled(True)
            self.ui.pinToolButton.setEnabled(True)

            self.followIpnsAction.setEnabled(
                self.currentIpfsObject.isIpns)
            self.curObjectCtrl.show()
        elif url.authority() == self.gatewayAuthority and 0:
            # Old code (when the dweb scheme handler was still
            # going through the go-ipfs http gw)
            #
            # dweb:/ with IPFS gateway's authority
            # Content loaded from IPFS gateway, this is IPFS content

            urlString = url.toDisplayString(
                QUrl.RemoveAuthority | QUrl.RemoveScheme)

            self._currentIpfsObject = IPFSPath(urlString)

            if self.currentIpfsObject.valid:
                log.debug('Current IPFS object: {0}'.format(
                    repr(self.currentIpfsObject)))

                url = QUrl(self.currentIpfsObject.dwebUrl)
                self.urlZone.clear()
                self.urlZone.insert(url.toString())

                self.ui.hashmarkThisPage.setEnabled(True)
                self.ui.pinToolButton.setEnabled(True)

                self.ipfsObjectVisited.emit(self.currentIpfsObject)

                self._currentUrl = url
            else:
                log.debug('Invalid IPFS path: {0}'.format(urlString))

            # Activate the follow action if this is a root IPNS address
            self.followIpnsAction.setEnabled(
                self.currentIpfsObject.isIpnsRoot)
            self.curObjectCtrl.show()
        else:
            if sHandler and issubclass(
                    sHandler.__class__, IPFSObjectProxyScheme):
                proxiedPath = await sHandler.urlProxiedPath(url)

                if proxiedPath and proxiedPath.valid:
                    self._currentIpfsObject = proxiedPath
                    self.ipfsObjectVisited.emit(self.currentIpfsObject)
                    self.ui.hashmarkThisPage.setEnabled(True)
                    self.curObjectCtrl.show()
            else:
                # Non-IPFS browsing
                self._currentIpfsObject = None
                self.curObjectCtrl.hide()

            self.urlZone.clear()
            self.urlZone.insert(url.toString())
            self._currentUrl = url
            self.followIpnsAction.setEnabled(False)

        currentPage = self.webView.page()

        if currentPage:
            history = currentPage.history()
            self.ui.backButton.setEnabled(history.canGoBack())
            self.ui.forwardButton.setEnabled(history.canGoForward())

    def onTitleChanged(self, pageTitle):
        if pageTitle.startswith(self.gatewayAuthority):
            pageTitle = iNoTitle()

        self.currentPageTitle = pageTitle

        lenMax = 16
        if len(pageTitle) > lenMax:
            pageTitle = '{0} ...'.format(pageTitle[0:lenMax])

        idx = self.gWindow.tabWidget.indexOf(self)
        self.gWindow.tabWidget.setTabText(idx, pageTitle)
        self.gWindow.tabWidget.setTabToolTip(idx, self.currentPageTitle)

        if self.currentPageTitle != iNoTitle() and self.currentUrl and \
                self.currentUrl.isValid():
            if isSchemeRegistered(self.currentUrl.scheme()):
                self.history.record(self.currentUrl.toString(),
                                    self.currentPageTitle)

    def onLoadFinished(self, ok):
        self.ui.stopButton.setEnabled(False)

    def onIconChanged(self, icon):
        self.gWindow.tabWidget.setTabIcon(self.tabPageIdx, icon)

    def onLoadProgress(self, progress):
        self.ui.pBarBrowser.setValue(progress)
        self.ui.stopButton.setEnabled(progress >= 0 and progress < 100)

        if progress == 100:
            self.ui.pBarBrowser.setStyleSheet(
                '''QProgressBar::chunk#pBarBrowser {
                    background-color: #244e66;
                }''')
            self.loop.call_later(
                1,
                self.ui.pBarBrowser.setStyleSheet,
                '''QProgressBar::chunk#pBarBrowser {
                    background-color: transparent;
                }''')
        else:
            self.ui.pBarBrowser.setStyleSheet(
                '''QProgressBar::chunk#pBarBrowser {
                    background-color: #4b9fa2;
                }''')

    def browseFsPath(self, path):
        def _handle(iPath):
            if iPath.valid:
                self.enterUrl(QUrl(iPath.ipfsUrl))
            else:
                messageBox(iInvalidObjectPath(iPath.fullPath))

        if isinstance(path, str):
            ipfsPath = IPFSPath(path, autoCidConv=True)
            return _handle(ipfsPath)
        elif isinstance(path, IPFSPath):
            return _handle(path)

    def browseIpfsHash(self, ipfsHash):
        if not cidhelpers.cidValid(ipfsHash):
            return messageBox(iInvalidCID(ipfsHash))

        self.browseFsPath(IPFSPath(joinIpfs(ipfsHash), autoCidConv=True))

    def browseIpnsKey(self, ipnsHash):
        self.browseFsPath(IPFSPath(joinIpns(ipnsHash)))

    def enterUrl(self, url):
        self.urlZone.hideMatches()
        self.urlZone.cancelTimer()

        if not url.isValid() or url.scheme() in ['data', SCHEME_Z]:
            messageBox('Invalid URL: {0}'.format(url.toString()))
            return

        self.urlZone.urlEditing = False
        self.urlZone.clearFocus()

        log.debug('Entering URL {}'.format(url.toString()))

        handler = self.webEngineView.webProfile.urlSchemeHandler(
            url.scheme().encode())

        def onEnsTimeout():
            pass

        # This should not be necessary starting with qt 5.13
        if not url.path():
            url.setPath('/')

        if url.scheme() == SCHEME_ENS:
            self.resolveTimer.timeout.connect(onEnsTimeout)
            self.resolveTimer.start(self.resolveTimerEnsMs)

        if handler:
            if hasattr(handler, 'webProfileNeeded'):
                wpName = handler.webProfileNeeded
                if self.webEngineView.webProfile.profileName != wpName and \
                        wpName in self.app.webProfiles:
                    self.changeWebProfileByName(wpName)

        self.urlZone.clear()
        self.urlZone.insert(url.toString())
        self.webEngineView.load(url)

    def handleEditedUrl(self, inputStr):
        self.urlZone.hideMatches()
        self.urlZone.cancelTimer()
        self.urlZone.unfocus()

        #
        # Handle seamless upgrade of CIDv0, suggested by @lidel
        #
        # If the user uses the native scheme (ipfs://) but passes
        # a base58-encoded CID as host (whatever the version),
        # convert it to base32 with cidhelpers.cidConvertBase32()
        # and replace the old CID in the URL
        #
        # https://github.com/eversum/galacteek/issues/5
        #

        if inputStr.startswith('ipfs://'):
            # Home run
            match = cidhelpers.ipfsDedSearchPath58(inputStr)
            if match:
                rootcid = match.group('rootcid')
                if rootcid and re.search('[A-Z]', rootcid):
                    # Has uppercase characters .. Either CIDv0
                    # or base58-encoded CIDv1
                    # If the conversion to base32 is successfull,
                    # replace the base58-encoded CID in the URL with
                    # the base32-encoded CIDv1

                    cid = cidhelpers.cidConvertBase32(rootcid)
                    if cid:
                        inputStr = re.sub(rootcid, cid, inputStr,
                                          count=1)
                    else:
                        return messageBox(iInvalidCID(rootcid))

        url = QUrl(inputStr)
        if not url.isValid() or not url.scheme():
            # Invalid URL or no scheme given
            # If the address bar contains a valid CID or IPFS path, load it
            iPath = IPFSPath(inputStr, autoCidConv=True)

            if iPath.valid:
                return self.browseFsPath(iPath)
            else:
                return messageBox(iInvalidUrl(inputStr))

        scheme = url.scheme()

        if isSchemeRegistered(scheme):
            self.enterUrl(url)
        elif scheme in ['http', 'https'] and \
                self.app.settingsMgr.allowHttpBrowsing is True:
            # Browse http urls if allowed
            self.enterUrl(url)
        else:
            messageBox('Unknown URL type')

    def onEnsResolved(self, domain, path):
        logUser.info('ENS: {0} maps to {1}'.format(domain, path))

    def onZoomIn(self):
        cFactor = self.webView.zoomFactor()
        self.webView.setZoomFactor(cFactor + 0.25)

    def onZoomOut(self):
        cFactor = self.webView.zoomFactor()
        self.webView.setZoomFactor(cFactor - 0.25)



import json

import asyncio
import time
import collections

from galacteek import log as logger

from galacteek.ipfs.pubsub import TOPIC_MAIN
from galacteek.ipfs.pubsub import TOPIC_PEERS
from galacteek.ipfs.pubsub import TOPIC_CHAT
from galacteek.ipfs.pubsub import TOPIC_HASHMARKS

from galacteek.ipfs.pubsub.messages import MarksBroadcastMessage
from galacteek.ipfs.pubsub.messages import PeerIdentMessageV1
from galacteek.ipfs.pubsub.messages import PeerIdentMessageV2
from galacteek.ipfs.pubsub.messages import PeerLogoutMessage
from galacteek.ipfs.pubsub.messages import ChatRoomMessage

from galacteek.ipfs.wrappers import ipfsOp
from galacteek.core.asynclib import asyncify
from galacteek.core.ipfsmarks import IPFSMarks

import aioipfs


class PubsubService(object):
    """
    Generic IPFS pubsub service
    """

    def __init__(self, ipfsCtx, client, topic='galacteek.default',
                 runPeriodic=False, filterSelfMessages=True,
                 maxMsgTsDiff=None, minMsgTsDiff=None,
                 maxMessageSize=32768):
        self.client = client
        self.ipfsCtx = ipfsCtx
        self.topic = topic
        self.inQueue = asyncio.Queue()
        self.errorsQueue = asyncio.Queue()
        self.lock = asyncio.Lock()

        self._receivedCount = 0
        self._errorsCount = 0
        self._runPeriodic = runPeriodic
        self._filters = []
        self._peerActivity = {}
        self._maxMsgTsDiff = maxMsgTsDiff
        self._minMsgTsDiff = minMsgTsDiff
        self._maxMessageSize = maxMessageSize

        self.tskServe = None
        self.tskProcess = None
        self.tskPeriodic = None

        self.addMessageFilter(self.filterMessageSize)
        self.addMessageFilter(self.filterPeerActivity)

        if filterSelfMessages is True:
            self.addMessageFilter(self.filterSelf)

    @property
    def receivedCount(self):
        return self._receivedCount

    @property
    def errorsCount(self):
        return self._errorsCount

    @property
    def runPeriodic(self):
        return self._runPeriodic

    def debug(self, msg):
        logger.debug('PS[{0}]: {1}'.format(self.topic, msg))

    def logStatus(self):
        self.debug('** Messages received: {0}, errors: {1}'.format(
            self.receivedCount, self.errorsCount))

    def start(self):
        """ Create the different tasks for this service """
        self.tskServe = self.ipfsCtx.loop.create_task(self.serve())
        self.tskProcess = self.ipfsCtx.loop.create_task(self.processMessages())

        if self.runPeriodic:
            self.tskPeriodic = self.ipfsCtx.loop.create_task(self.periodic())

    async def stop(self):
        await self.shutdown()
        for tsk in [self.tskServe, self.tskProcess, self.tskPeriodic]:
            if not tsk:
                continue

            tsk.cancel()
            try:
                await tsk
            except asyncio.CancelledError as cErr:
                self.debug('task {task}: shutdown ERR: {err}'.format(
                    task=tsk, err=str(cErr)))
                continue
            else:
                self.debug('task {}: shutdown ok'.format(tsk))

    def addMessageFilter(self, filtercoro):
        if asyncio.iscoroutinefunction(filtercoro):
            self._filters.append(filtercoro)

    async def filterMessageSize(self, msg):
        if len(msg['data']) > self._maxMessageSize:
            return True
        return False

    async def filterPeerActivity(self, msg):
        senderNodeId = msg['from'].decode()
        if senderNodeId not in self._peerActivity:
            self._peerActivity[senderNodeId] = collections.deque([], 10)

        records = self._peerActivity[senderNodeId]
        records.append((time.time(), len(msg['data'])))

        if len(records) > 2 and isinstance(self._minMsgTsDiff, int):
            latestIdx = len(records) - 1
            latest = records[latestIdx][0]
            previous = records[latestIdx - 1][0]

            if (latest - previous) < self._minMsgTsDiff:
                return True

        return False

    async def filterSelf(self, msg):
        if msg['from'].decode() == self.ipfsCtx.node.id:
            return True

        return False

    async def filtered(self, message):
        for filter in self._filters:
            if await filter(message):  # That means we drop it
                return True
        return False

    async def shutdown(self):
        pass

    async def processMessages(self):
        """ Messages processing task to be implemented in your listener """
        return True

    async def periodic(self):
        return True

    async def serve(self):
        """
        Subscribe to the pubsub topic, read and filter incoming messages
        (dropping the ones emitted by us). Selected messages are put on the
        async queue (inQueue) to be later treated in the processMessages() task
        """

        try:
            async for message in self.client.pubsub.sub(self.topic):
                if await self.filtered(message):
                    continue

                self.ipfsCtx.pubsub.psMessageRx.emit()

                await self.inQueue.put(message)
                self._receivedCount += 1
                await asyncio.sleep(0)
        except asyncio.CancelledError:
            self.debug('Cancelled, queue size was {0}'.format(
                self.inQueue.qsize()))
            return
        except Exception as e:
            self.debug('Serve interrupted by unknown exception {}'.format(
                str(e)))

    async def send(self, data, topic=None):
        """
        Publish a message

        :param str data: message payload
        """

        if topic is None:
            topic = self.topic

        try:
            status = await self.client.pubsub.pub(topic, data)
            self.ipfsCtx.pubsub.psMessageTx.emit()
        except aioipfs.APIError:
            logger.debug('Could not send pubsub message to {topic}'.format(
                topic=topic))
        else:
            return status


class JSONPubsubService(PubsubService):
    """
    JSON pubsub listener, handling incoming messages as JSON objects
    """

    def msgDataToJson(self, msg):
        """
        Decode JSON data contained in a pubsub message
        """
        try:
            return json.loads(msg['data'].decode())
        except Exception:
            logger.debug('Could not decode JSON message data')
            return None

    async def processMessages(self):
        try:
            while True:
                data = await self.inQueue.get()

                if data is None:
                    break

                msg = self.msgDataToJson(data)
                if msg is None:
                    continue

                try:
                    await self.processJsonMessage(data['from'].decode(), msg)
                except Exception as exc:
                    logger.debug(
                        'processJsonMessage error: {}'.format(str(exc)))
                    await self.errorsQueue.put((msg, exc))
                    self._errorsCount += 1
        except asyncio.CancelledError:
            return
        except Exception:
            return

    async def processJsonMessage(self, sender, msg):
        """ Implement this method to process incoming JSON messages"""
        return True


class PSHashmarksExchanger(JSONPubsubService):
    def __init__(self, ipfsCtx, client, marksLocal, marksNetwork):
        super().__init__(
            ipfsCtx,
            client,
            topic=TOPIC_HASHMARKS,
            runPeriodic=True,
            filterSelfMessages=True,
            maxMessageSize=131072,
            minMsgTsDiff=60 * 2)
        self.marksLocal = marksLocal
        self.marksNetwork = marksNetwork
        self.marksLocal.markAdded.connect(self.onMarkAdded)
        self._sendEvery = 60 * 8
        self._lastBroadcast = 0

    @property
    def curProfile(self):
        return self.ipfsCtx.currentProfile

    @asyncify
    async def onMarkAdded(self, path, mark):
        if mark['share'] is True:
            await self.broadcastSharedMarks()

    async def broadcastSharedMarks(self):
        try:
            sharedMarks = IPFSMarks(None, autosave=False)
            count = sharedMarks.merge(self.marksLocal, share=True, reset=True)

            if count > 0:
                self.debug('Sending shared hashmarks')
                msg = MarksBroadcastMessage.make(self.ipfsCtx.node.id,
                                                 sharedMarks.root)
                await self.send(str(msg))
                self._lastBroadcast = int(time.time())
        except BaseException:
            pass

    async def periodic(self):
        while True:
            await asyncio.sleep(self._sendEvery)
            await self.broadcastSharedMarks()

    async def processJsonMessage(self, sender, msg):
        msgType = msg.get('msgtype', None)
        if msgType == MarksBroadcastMessage.TYPE:
            try:
                await self.processBroadcast(sender, msg)
            except BaseException:
                pass

    async def processBroadcast(self, sender, msg):
        bMsg = MarksBroadcastMessage(msg)

        if not bMsg.valid():
            self.debug('Invalid broadcast message')
            return

        if not self.ipfsCtx.peers.peerRegistered(sender):
            self.debug('Broadcast from unregistered peer: {0}'.format(sender))
            return

        marksJson = bMsg.marks
        if not marksJson:
            return

        marksCollection = IPFSMarks(None, data=marksJson, autosave=False)

        # JSON schema validation
        if not await marksCollection.isValidAsync():
            self.debug('Received invalid hashmarks from {sender}'.format(
                sender=sender))
            return
        else:
            self.debug('Hashmarks broadcast from {sender} is valid'.format(
                sender=sender))

        try:
            if self.curProfile:
                await self.curProfile.storeHashmarks(sender, marksCollection)
        except Exception:
            self.debug('Could not store data in hashmarks library')
        else:
            self.info('Stored hashmarks from peer: {0}'.format(sender))


class PSMainService(JSONPubsubService):
    def __init__(self, ipfsCtx, client):
        super().__init__(ipfsCtx, client, topic=TOPIC_MAIN)


class PSPeersService(JSONPubsubService):
    def __init__(self, ipfsCtx, client):
        super().__init__(ipfsCtx, client, topic=TOPIC_PEERS,
                         runPeriodic=True)

        self._curProfile = None
        self._identEvery = 45
        self.ipfsCtx.profileChanged.connect(self.onProfileChanged)

    @property
    def curProfile(self):
        return self.ipfsCtx.currentProfile

    @property
    def identEvery(self):
        return self._identEvery

    @asyncify
    async def onProfileChanged(self, pName, profile):
        await profile.userInfo.loaded

        try:
            profile.userInfo.entryChanged.disconnect(self.userInfoChanged)
            profile.userInfo.available.disconnect(self.userInfoAvail)
        except Exception:
            pass

        self.curProfile.userInfo.entryChanged.connect(self.userInfoChanged)
        self.curProfile.userInfo.available.connect(self.userInfoAvail)
        await self.sendIdent(self.curProfile)

    @asyncify
    async def userInfoAvail(self, arg):
        await self.sendIdent(self.curProfile)

    @asyncify
    async def userInfoChanged(self):
        with await self.lock:
            await self.sendIdent(self.curProfile)

    @ipfsOp
    async def sendIdent(self, op, profile):
        if not profile.initialized:
            logger.debug('Profile not initialized, ident message not sent')
            return

        nodeId = op.ctx.node.id
        uInfo = profile.userInfo

        if not uInfo.valid():
            logger.debug('Profile info, ident message not sent')
            return

        cfgMaps = []

        if self.ipfsCtx.inOrbit:
            if profile.orbitalCfgMap:
                cfgMaps.append(profile.orbitalCfgMap.data)

        if uInfo.schemaVersion == 1:
            msg = PeerIdentMessageV2.make(
                nodeId,
                uInfo.objHash,
                uInfo.root,
                profile.dagUser.dagCid,
                profile.keyRootId,
                self.ipfsCtx.p2p.servicesFormatted(),
                cfgMaps
            )

            logger.debug('Sending ident message')
            await self.send(str(msg))
        else:
            logger.debug('Unknown schema version, ident message not sent')

    async def processJsonMessage(self, sender, msg):
        msgType = msg.get('msgtype', None)

        if msgType == PeerIdentMessageV1.TYPE:
            logger.debug('Received ident message (v1) from {}'.format(sender))
            await self.handleIdentMessageV1(sender, msg)
        elif msgType == PeerIdentMessageV2.TYPE:
            logger.debug('Received ident message (v2) from {}'.format(sender))
            await self.handleIdentMessageV2(sender, msg)
        elif msgType == PeerLogoutMessage.TYPE:
            logger.debug('Received logout message from {}'.format(sender))
            await self.handleLogoutMessage(sender, msg)

        await asyncio.sleep(0)

    async def handleLogoutMessage(self, sender, msg):
        lMsg = PeerLogoutMessage(msg)
        if lMsg.valid():
            await self.ipfsCtx.peers.unregister(sender)

    async def handleIdentMessageV1(self, sender, msg):
        iMsg = PeerIdentMessageV1(msg)
        if not iMsg.valid():
            logger.debug('Received invalid ident message')
            return

        if sender != iMsg.peer:
            # You forging pubsub messages, son ?
            return

        await self.ipfsCtx.peers.registerFromIdent(iMsg)

    async def handleIdentMessageV2(self, sender, msg):
        iMsg = PeerIdentMessageV2(msg)
        if not iMsg.valid():
            logger.debug('Received invalid ident message')
            return

        if sender != iMsg.peer:
            # You forging pubsub messages, son ?
            return

        await self.ipfsCtx.peers.registerFromIdent(iMsg)

    @ipfsOp
    async def shutdown(self, op):
        msg = PeerLogoutMessage.make(op.ctx.node.id)
        await self.send(str(msg))

    async def periodic(self):
        while True:
            await asyncio.sleep(self.identEvery)

            with await self.lock:
                if self.curProfile:
                    await self.sendIdent(self.curProfile)


class PSChatService(JSONPubsubService):
    def __init__(self, ipfsCtx, client):
        super().__init__(ipfsCtx, client, topic=TOPIC_CHAT,
                         filterSelfMessages=False)

    async def processJsonMessage(self, sender, msg):
        msgType = msg.get('msgtype', None)

        if msgType == ChatRoomMessage.TYPE:
            await self.handleChatMessage(msg)

    async def handleChatMessage(self, msg):
        cMsg = ChatRoomMessage(msg)
        if not cMsg.valid():
            return

        self.ipfsCtx.pubsub.chatRoomMessageReceived.emit(cMsg)


__all__ = [
    'PubsubService',
    'PSHashmarksExchanger',
    'PSMainService',
    'PSChatService',
    'PSPeersService'
]



import functools
import os.path

from logbook import Handler
from logbook import StringFormatterHandlerMixin

from PyQt5.QtWidgets import QTabWidget
from PyQt5.QtWidgets import QMainWindow
from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QToolBar
from PyQt5.QtWidgets import QMenu
from PyQt5.QtWidgets import QAction
from PyQt5.QtWidgets import QActionGroup
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QTextBrowser
from PyQt5.QtWidgets import QToolTip

from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QTimer
from PyQt5.QtCore import QDateTime
from PyQt5.QtCore import QSize
from PyQt5.QtCore import QPoint
from PyQt5.QtCore import QRect

from PyQt5.Qt import QSizePolicy

from PyQt5.QtGui import QKeySequence
from PyQt5.QtGui import QTextCursor

from PyQt5 import QtWebEngineWidgets

from galacteek import GALACTEEK_NAME
from galacteek import ensure
from galacteek import log
from galacteek.core.glogger import loggerMain
from galacteek.core.glogger import loggerUser
from galacteek.core.glogger import easyFormatString
from galacteek.core.asynclib import asyncify
from galacteek.ui import mediaplayer
from galacteek.ipfs.wrappers import *
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.cidhelpers import joinIpns

from galacteek.core.orbitdb import GalacteekOrbitConnector
from galacteek.dweb.page import HashmarksPage, DWebView, WebTab
from galacteek.core.modelhelpers import *

from . import ui_ipfsinfos

from . import userwebsite
from . import browser
from . import files
from . import keys
from . import settings
from . import orbital
from . import textedit
from . import ipfsview
from . import ipfssearch
from . import peers
from . import eventlog
from . import pin
from . import chat

from .clips import RotatingCubeClipSimple
from .clips import RotatingCubeRedFlash140d
from .eth import EthereumStatusButton
from .feeds import AtomFeedsViewTab
from .feeds import AtomFeedsView
from .textedit import TextEditorTab
from .pyramids import MultihashPyramidsToolBar
from .quickaccess import QuickAccessToolBar
from .helpers import *
from .widgets import PopupToolButton
from .widgets import HashmarkMgrButton
from .widgets import HashmarksSearcher
from .widgets import AtomFeedsToolbarButton
from .widgets import AnimatedLabel
from .dialogs import *
from ..appsettings import *
from .i18n import *

from .clipboard import ClipboardManager
from .clipboard import ClipboardItemsStack


def iPinningItemStatus(pinPath, pinProgress):
    return QCoreApplication.translate(
        'GalacteekWindow',
        '\nPath: {0}, nodes processed: {1}').format(pinPath, pinProgress)


def iAbout():
    from galacteek import __version__
    return QCoreApplication.translate('GalacteekWindow', '''
        <p align='center'>
        <img src=':/share/icons/galacteek-incandescent.png' />
        </p>

        <p>
        <b>galacteek</b> is a multi-platform Qt5-based browser
        for the distributed web
        </p>
        <br/>
        <p>Contact:
            <a href="mailto: galacteek@protonmail.com">
                galacteek@protonmail.com
            </a>
        </p>
        <p>Authors: see
        <a href="https://github.com/eversum/galacteek/blob/master/AUTHORS.rst">
            AUTHORS.rst
        </a>
        </p>
        <p>galacteek version {0}</p>''').format(__version__)


class UserLogsWindow(QMainWindow):
    hidden = pyqtSignal()

    def hideEvent(self, event):
        self.hidden.emit()
        super().hideEvent(event)


class MainWindowLogHandler(Handler, StringFormatterHandlerMixin):
    """
    Custom logbook handler that logs to the status bar

    Should be moved to a separate module
    """

    modulesColorTable = {
        'galacteek.ui.resource': '#7f8491',
        'galacteek.core.profile': 'blue'
    }

    def __init__(self, logsBrowser, application_name=None, address=None,
                 facility='user', level=0, format_string=None,
                 filter=None, bubble=True, window=None):
        Handler.__init__(self, level, filter, bubble)
        StringFormatterHandlerMixin.__init__(self, easyFormatString)
        self.application_name = application_name
        self.window = window
        self.logsBrowser = logsBrowser

    def emit(self, record):
        fRecord = self.format(record)

        if record.level_name == 'INFO':
            color = self.modulesColorTable.get(record.module, 'black')
            self.window.statusMessage(
                "<p style='color: {color}'>{msg}</p>\n".format(
                    color=color, msg=fRecord))

        self.logsBrowser.append(fRecord)

        if not self.logsBrowser.isVisible():
            self.logsBrowser.moveCursor(QTextCursor.End)


class IPFSInfosDialog(QDialog):
    """
    IPFS node/repository information dialog
    """

    def __init__(self, app, parent=None):
        super().__init__(parent)

        self.ui = ui_ipfsinfos.Ui_IPFSInfosDialog()
        self.ui.setupUi(self)
        self.ui.okButton.clicked.connect(functools.partial(self.done, 1))

        self.labels = [self.ui.repoObjCount,
                       self.ui.repoVersion,
                       self.ui.repoSize,
                       self.ui.repoMaxStorage,
                       self.ui.nodeId,
                       self.ui.agentVersion,
                       self.ui.protocolVersion]

        self.enableLabels(False)

    def enableLabels(self, enable):
        for label in self.labels:
            if not enable:
                label.setText('Fetching ...')

            label.setEnabled(enable)

    @ipfsOp
    async def loadInfos(self, ipfsop):
        try:
            repoStat = await ipfsop.client.repo.stat()
            idInfo = await ipfsop.client.core.id()
        except BaseException:
            for label in self.labels:
                label.setText(iUnknown())
        else:
            self.ui.repoObjCount.setText(str(repoStat.get(
                'NumObjects', iUnknown())))
            self.ui.repoVersion.setText(str(repoStat.get(
                'Version', iUnknown())))
            self.ui.repoSize.setText(sizeFormat(repoStat.get(
                'RepoSize', 0)))
            self.ui.repoMaxStorage.setText(sizeFormat(repoStat.get(
                'StorageMax', 0)))

            self.ui.nodeId.setText(idInfo.get('ID', iUnknown()))
            self.ui.agentVersion.setText(idInfo.get(
                'AgentVersion', iUnknown()))
            self.ui.protocolVersion.setText(idInfo.get('ProtocolVersion',
                                                       iUnknown()))
            self.enableLabels(True)


class DatabasesManager(QObject):
    def __init__(self, orbitConnector, parent=None):
        super().__init__(parent)

        self.mainW = parent

        self.icon = getIcon('orbitdb.png')
        self.connector = orbitConnector
        self._dbButton = self.buildButton()

    @property
    def button(self):
        return self._dbButton

    def buildButton(self):
        dbButton = QToolButton(self)
        dbButton.setIconSize(QSize(24, 24))
        dbButton.setIcon(self.icon)
        dbButton.setPopupMode(QToolButton.MenuButtonPopup)

        self.databasesMenu = QMenu(self)
        self.mainFeedAction = QAction(self.icon,
                                      'General discussions feed', self,
                                      triggered=self.onMainFeed)
        self.databasesMenu.addAction(self.mainFeedAction)
        dbButton.setMenu(self.databasesMenu)
        return dbButton

    def onMainFeed(self):
        database = self.connector.database('feeds', 'general')
        view = orbital.OrbitFeedView(self.connector, database,
                                     parent=self.mainW.tabWidget)
        self.mainW.registerTab(view, 'General', current=True,
                               icon=self.icon)


class MiscToolBar(QToolBar):
    moved = pyqtSignal(int)

    def __init__(self, parent):
        super().__init__(parent)

        self.setObjectName('toolbarMisc')
        self.setAllowedAreas(
            Qt.RightToolBarArea
        )
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def contextMenuEvent(self, event):
        # no context menu
        pass


class MainToolBar(QToolBar):
    moved = pyqtSignal(int)

    def __init__(self, parent):
        super().__init__(parent)
        self.setObjectName('toolbarMain')

        self.setFloatable(False)
        self.setAllowedAreas(
            Qt.LeftToolBarArea | Qt.TopToolBarArea
        )
        self.setContextMenuPolicy(Qt.NoContextMenu)

        # Empty widget
        self.emptySpace = QWidget()
        self.emptySpace.setSizePolicy(
            QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.lastPos = None

    @property
    def vertical(self):
        return self.orientation() == Qt.Vertical

    @property
    def horizontal(self):
        return self.orientation() == Qt.Horizontal

    def dragEnterEvent(self, event):
        event.acceptProposedAction()

    def dropEvent(self, event):
        pass


class ProfileButton(PopupToolButton):
    pass


class TabWidgetKeyFilter(QObject):
    nextPressed = pyqtSignal()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress:
            modifiers = event.modifiers()

            key = event.key()
            if modifiers & Qt.ControlModifier:
                if key == Qt.Key_J:
                    self.nextPressed.emit()
                    return True
        return False


class CentralWidget(QWidget):
    def __init__(self, parent):
        super(CentralWidget, self).__init__(parent)

        self.setObjectName('centralWidget')
        self.wLayout = QVBoxLayout()
        self.setLayout(self.wLayout)


class BrowseButton(PopupToolButton):
    """
    Browse button. When the button is hovered we play the
    rotating cube clip.
    """

    def __init__(self, parent):
        super().__init__(parent=parent, mode=QToolButton.InstantPopup)

        self.animatedActions = []
        self.rotatingCubeClip = RotatingCubeClipSimple()
        self.rotatingCubeClip.finished.connect(
            functools.partial(self.rotatingCubeClip.start))
        self.rotatingCubeClip.frameChanged.connect(self.onCubeClipFrame)
        self.normalIcon()

    def setMenu(self, menu):
        super(BrowseButton, self).setMenu(menu)

        menu.triggered.connect(lambda action: self.normalIcon())
        menu.aboutToHide.connect(self.normalIcon)

    def onCubeClipFrame(self, no):
        icon = self.rotatingCubeClip.createIcon()

        self.setIcon(icon)

        if self.menu and self.menu.isVisible():
            for action in self.menu.actions():
                if action in self.animatedActions:
                    action.setIcon(icon)

    def rotateCube(self):
        if not self.rotatingCubeClip.playing():
            self.rotatingCubeClip.start()

    def normalIcon(self):
        self.rotatingCubeClip.stop()
        self.setIcon(getIconIpfs64())

    def enterEvent(self, event):
        self.rotateCube()
        super(BrowseButton, self).enterEvent(event)

    def leaveEvent(self, event):
        if not self.menu.isVisible() and self.isEnabled():
            self.normalIcon()

        super(BrowseButton, self).leaveEvent(event)


class MainWindow(QMainWindow):
    def __init__(self, app):
        super(MainWindow, self).__init__()

        self.showMaximized()
        self._app = app
        self._allTabs = []
        self._lastFeedMark = None

        self.centralWidget = CentralWidget(self)
        self.menuBar().hide()

        # Seems reasonable
        self.setMinimumSize(QSize(
            self.app.desktopGeometry.width() / 2,
            self.app.desktopGeometry.height() / 2)
        )

        # User logs widget
        self.logsPopupWindow = UserLogsWindow()
        self.logsPopupWindow.setObjectName('logsTextWindow')
        self.logsPopupWindow.setWindowTitle('{}: logs'.format(GALACTEEK_NAME))
        self.logsPopupWindow.hide()

        self.logsBrowser = QTextBrowser(self.logsPopupWindow)
        self.logsBrowser.setObjectName('logsTextWidget')
        self.logsPopupWindow.setCentralWidget(self.logsBrowser)
        self.logsBrowser.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.logsPopupWindow.setMinimumSize(QSize(
            (2 * self.app.desktopGeometry.width()) / 3,
            self.app.desktopGeometry.height() / 2
        ))

        self.userLogsHandler = MainWindowLogHandler(
            self.logsBrowser, window=self,
            level='DEBUG' if self.app.debugEnabled else 'INFO')

        loggerMain.handlers.append(self.userLogsHandler)
        loggerUser.handlers.append(self.userLogsHandler)

        self.tabnFManager = iFileManager()
        self.tabnKeys = iKeys()
        self.tabnPinning = iPinningStatus()
        self.tabnMediaPlayer = iMediaPlayer()
        self.tabnHashmarks = iHashmarks()
        self.tabnChat = iChat()
        self.tabnFeeds = iAtomFeeds()

        self.menuManual = QMenu(iManual(), self)

        # Global pin-all button
        self.pinAllGlobalButton = QToolButton(self)
        self.pinAllGlobalButton.setIcon(getIcon('pin.png'))
        self.pinAllGlobalButton.setObjectName('pinGlobalButton')
        self.pinAllGlobalButton.setToolTip(iGlobalAutoPinning())
        self.pinAllGlobalButton.setCheckable(True)
        self.pinAllGlobalButton.setAutoRaise(True)
        self.pinAllGlobalChecked = False
        self.pinAllGlobalButton.toggled.connect(self.onToggledPinAllGlobal)
        self.pinAllGlobalButton.setChecked(self.app.settingsMgr.browserAutoPin)

        self.toolbarMain = MainToolBar(self)
        self.toolbarPyramids = MultihashPyramidsToolBar(self)

        self.toolbarMain.orientationChanged.connect(self.onMainToolbarMoved)
        self.toolbarTools = QToolBar()
        self.toolbarTools.setOrientation(self.toolbarMain.orientation())

        # Apps/shortcuts toolbar
        self.qaToolbar = QuickAccessToolBar(self)
        self.qaToolbar.setOrientation(self.toolbarMain.orientation())

        # Main actions and browse button setup
        self.quitAction = QAction(getIcon('quit.png'),
                                  iQuit(),
                                  shortcut=QKeySequence('Ctrl+q'),
                                  triggered=self.quit)
        self.quitAction.setShortcutVisibleInContextMenu(True)

        self.restartAction = QAction(getIcon('quit.png'),
                                     iRestart(),
                                     shortcut=QKeySequence('Alt+Shift+r'),
                                     triggered=self.app.restart)

        self.mPlayerOpenAction = QAction(getIcon('multimedia.png'),
                                         iMediaPlayer(),
                                         triggered=self.onOpenMediaPlayer)

        self.editorOpenAction = QAction(getIcon('text-editor.png'),
                                        iTextEditor(),
                                        triggered=self.addEditorTab)

        self.hashmarksManagerAction = QAction(getIcon('hashmarks.png'),
                                              iHashmarksManager(),
                                              shortcut=QKeySequence('Ctrl+m'),
                                              triggered=self.addHashmarksTab)

        self.editorOpenAction = QAction(getIcon('text-editor.png'),
                                        iTextEditor(),
                                        triggered=self.addEditorTab)

        self.browseAction = QAction(
            getIconIpfsIce(),
            iBrowse(),
            shortcut=QKeySequence('Ctrl+t'),
            triggered=self.onOpenBrowserTabClicked)
        self.browseAction.setShortcutVisibleInContextMenu(True)
        self.browseAutopinAction = QAction(
            getIconIpfsIce(),
            iBrowseAutoPin(),
            shortcut=QKeySequence('Ctrl+Alt+t'),
            triggered=functools.partial(self.onOpenBrowserTabClicked,
                                        pinBrowsed=True))
        self.browseAutopinAction.setShortcutVisibleInContextMenu(True)

        self.chatAction = QAction(
            getIcon('chat.png'),
            'Chat',
            triggered=self.onOpenChatWidget)

        self.browseButton = BrowseButton(self)
        self.browseButton.setPopupMode(QToolButton.InstantPopup)
        self.browseButton.setObjectName('buttonBrowseIpfs')
        self.browseButton.normalIcon()

        self.browseButton.menu.addAction(self.browseAction)
        self.browseButton.menu.addAction(self.browseAutopinAction)
        self.browseButton.menu.addSeparator()
        self.browseButton.menu.addAction(self.editorOpenAction)
        self.browseButton.menu.addSeparator()
        self.browseButton.menu.addAction(self.chatAction)
        self.browseButton.menu.addSeparator()
        self.browseButton.menu.addAction(self.mPlayerOpenAction)
        self.browseButton.menu.addSeparator()
        self.browseButton.menu.addAction(self.quitAction)

        self.browseButton.animatedActions = [
            self.browseAction,
            self.browseAutopinAction
        ]
        self.browseButton.rotateCube()

        # File manager button
        self.fileManagerButton = QToolButton(self)
        self.fileManagerButton.setToolTip(iFileManager())
        self.fileManagerButton.setIcon(getIcon('folder-open.png'))
        self.fileManagerButton.clicked.connect(self.onFileManagerClicked)
        self.fileManagerButton.setShortcut(QKeySequence('Ctrl+Alt+f'))

        # File manager
        self.fileManagerWidget = files.FileManager(parent=self)

        # Text editor button
        self.textEditorButton = QToolButton(self)
        self.textEditorButton.setToolTip(iTextEditor())
        self.textEditorButton.setIcon(getIcon('text-editor.png'))
        self.textEditorButton.clicked.connect(self.addEditorTab)

        # Atom Feeds
        self.atomButton = AtomFeedsToolbarButton(self)
        self.atomButton.clicked.connect(self.onShowAtomFeeds)
        self.atomFeedsViewWidget = AtomFeedsView(self.app.modelAtomFeeds)

        # Edit-Profile button
        self.menuUserProfile = QMenu(self)
        self.profilesActionGroup = QActionGroup(self)

        # Profile button
        self.profileMenu = QMenu(self)
        iconProfile = getIcon('profile-user.png')
        self.profileMenu.addAction(iconProfile,
                                   'Edit profile',
                                   self.onProfileEditDialog)
        self.profileMenu.addAction(getIcon('go-home.png'),
                                   'View homepage',
                                   self.onProfileViewHomepage)
        self.profileMenu.addSeparator()

        self.userWebsiteManager = userwebsite.UserWebsiteManager(
            parent=self.profileMenu)
        self.profileMenu.addMenu(self.userWebsiteManager.blogMenu)

        self.profileEditButton = ProfileButton(
            menu=self.profileMenu,
            mode=QToolButton.InstantPopup,
            icon=iconProfile
        )
        self.profileEditButton.setEnabled(False)

        # Hashmarks mgr button

        # Shared hashmarks mgr button
        self.hashmarksSearcher = HashmarksSearcher(parent=self)
        self.hashmarksSearcher.setToolTip(iHashmarksLibrary())

        self.hashmarkMgrButton = HashmarkMgrButton(
            marks=self.app.marksLocal)
        self.hashmarkMgrButton.menu.addAction(self.hashmarksManagerAction)
        self.hashmarkMgrButton.menu.addSeparator()
        self.hashmarkMgrButton.menu.addMenu(self.hashmarksSearcher.menu)

        self.hashmarkMgrButton.menu.addSeparator()
        self.hashmarkMgrButton.updateMenu()

        # Peers button
        self.peersButton = QToolButton(self)
        self.peersButton.setIcon(getIcon('peers.png'))
        self.peersButton.clicked.connect(self.onPeersMgrClicked)

        self.clipboardItemsStack = ClipboardItemsStack(parent=self.toolbarMain)

        # Clipboard loader button
        self.clipboardManager = ClipboardManager(
            self.app.clipTracker,
            self.clipboardItemsStack,
            self.app.resourceOpener,
            icon=getIcon('clipboard.png'),
            parent=self.toolbarMain
        )

        # Settings button
        settingsIcon = getIcon('settings.png')
        self.settingsToolButton = QToolButton(self)
        self.settingsToolButton.setIcon(settingsIcon)
        self.settingsToolButton.setPopupMode(QToolButton.InstantPopup)
        menu = QMenu(self)
        menu.addAction(settingsIcon, iSettings(),
                       self.onSettings)
        menu.addSeparator()
        menu.addAction(settingsIcon, iEventLog(),
                       self.onOpenEventLog)
        menu.addAction(getIcon('lock-and-key.png'), iKeys(),
                       self.onIpfsKeysClicked)
        menu.addSeparator()
        menu.addAction(iClearHistory(), self.onClearHistory)

        self.settingsToolButton.setMenu(menu)

        # Help button
        self.helpToolButton = QToolButton(self)
        self.helpToolButton.setObjectName('helpToolButton')
        self.helpToolButton.setIcon(getIcon('information.png'))
        self.helpToolButton.setPopupMode(QToolButton.InstantPopup)
        menu = QMenu(self)
        menu.addMenu(self.menuManual)
        menu.addSeparator()
        menu.addAction('Donate', self.onHelpDonate)
        menu.addAction('About', self.onAboutGalacteek)
        self.helpToolButton.setMenu(menu)

        # Quit button
        self.quitButton = PopupToolButton(
            parent=self, mode=QToolButton.InstantPopup)
        self.quitButton.setObjectName('quitToolButton')
        self.quitButton.setIcon(self.quitAction.icon())
        self.quitButton.menu.addAction(self.restartAction)
        self.quitButton.menu.addSeparator()
        self.quitButton.menu.addAction(self.quitAction)
        self.quitButton.setToolTip(iQuit())

        self.ipfsSearchPageFactory = ipfssearch.SearchResultsPageFactory(self)

        self.ipfsSearchButton = ipfssearch.IPFSSearchButton(self)
        self.ipfsSearchButton.setShortcut(QKeySequence('Ctrl+Alt+s'))
        self.ipfsSearchButton.setIcon(self.ipfsSearchButton.iconNormal)
        self.ipfsSearchButton.clicked.connect(self.addIpfsSearchView)

        self.toolbarMain.addWidget(self.browseButton)
        self.toolbarMain.addWidget(self.hashmarkMgrButton)
        self.toolbarMain.addWidget(self.hashmarksSearcher)
        self.toolbarMain.addWidget(self.atomButton)

        self.toolbarMain.addSeparator()
        self.toolbarMain.addWidget(self.fileManagerButton)
        self.toolbarMain.addWidget(self.textEditorButton)

        self.hashmarkMgrButton.hashmarkClicked.connect(self.onHashmarkClicked)
        self.hashmarksSearcher.hashmarkClicked.connect(
            self.onHashmarkClicked)

        self.toolbarMain.addSeparator()

        self.toolbarMain.addWidget(self.toolbarTools)
        self.toolbarMain.addSeparator()
        self.toolbarMain.addWidget(self.qaToolbar)

        self.toolbarTools.addWidget(self.peersButton)
        self.toolbarTools.addWidget(self.profileEditButton)

        self.toolbarMain.actionStatuses = self.toolbarMain.addAction(
            'Statuses')
        self.toolbarMain.actionStatuses.setVisible(False)

        self.toolbarMain.addWidget(self.toolbarMain.emptySpace)
        self.toolbarMain.addSeparator()
        self.toolbarMain.addWidget(self.ipfsSearchButton)
        self.toolbarMain.addSeparator()
        self.toolbarMain.addWidget(self.clipboardItemsStack)
        self.toolbarMain.addWidget(self.clipboardManager)
        self.toolbarMain.addSeparator()

        self.toolbarMain.addWidget(self.pinAllGlobalButton)
        self.toolbarMain.addWidget(self.settingsToolButton)

        self.toolbarMain.addWidget(self.helpToolButton)
        self.toolbarMain.addSeparator()
        self.toolbarMain.addWidget(self.quitButton)

        self.addToolBar(Qt.TopToolBarArea, self.toolbarMain)
        self.addToolBar(Qt.RightToolBarArea, self.toolbarPyramids)

        self.tabWidget = QTabWidget(self)
        self.tabWidget.setObjectName('tabWidget')
        self.tabWidget.setTabsClosable(True)
        self.tabWidget.tabCloseRequested.connect(self.onTabCloseRequest)
        self.tabWidget.setElideMode(Qt.ElideMiddle)
        self.tabWidget.setUsesScrollButtons(True)

        self.centralWidget.wLayout.addWidget(self.tabWidget)

        if self.app.system != 'Darwin':
            self.tabWidget.setDocumentMode(True)

        tabKeyFilter = TabWidgetKeyFilter(self)
        tabKeyFilter.nextPressed.connect(self.cycleTabs)
        self.tabWidget.installEventFilter(tabKeyFilter)

        # Chat room
        self.chatRoomWidget = chat.ChatRoomWidget(self)

        self.webProfile = QtWebEngineWidgets.QWebEngineProfile.defaultProfile()

        # Status bar setup
        self.pinningStatusButton = QPushButton(self)
        self.pinningStatusButton.setShortcut(
            QKeySequence('Ctrl+u'))
        self.pinningStatusButton.setToolTip(iNoStatus())
        self.pinningStatusButton.setIcon(getIcon('pin-black.png'))
        self.pinningStatusButton.clicked.connect(
            self.showPinningStatusWidget)
        self.pubsubStatusButton = QPushButton(self)
        self.pubsubStatusButton.setIcon(getIcon('network-offline.png'))

        self.ipfsStatusCube = AnimatedLabel(
            RotatingCubeRedFlash140d(speed=10),
            parent=self
        )
        self.ipfsStatusCube.clip.setScaledSize(QSize(24, 24))
        self.ipfsStatusCube.animationClicked.connect(self.onIpfsInfos)
        self.ipfsStatusCube.startClip()

        self.ethereumStatusBtn = EthereumStatusButton(parent=self)

        self.statusbar = self.statusBar()
        self.userLogsButton = QToolButton(self)
        self.userLogsButton.setToolTip('Logs')
        self.userLogsButton.setIcon(getIcon('information.png'))
        self.userLogsButton.setCheckable(True)
        self.userLogsButton.toggled.connect(self.onShowUserLogs)
        self.logsPopupWindow.hidden.connect(
            functools.partial(self.userLogsButton.setChecked, False))

        self.lastLogLabel = QLabel(self.statusbar)
        self.lastLogLabel.setAlignment(Qt.AlignLeft)
        self.lastLogLabel.setObjectName('lastLogLabel')
        self.statusbar.insertWidget(0, self.lastLogLabel, 1)

        self.statusbar.addPermanentWidget(self.ipfsStatusCube)

        self.statusbar.addPermanentWidget(self.ethereumStatusBtn)
        self.statusbar.addPermanentWidget(self.pinningStatusButton)
        self.statusbar.addPermanentWidget(self.pubsubStatusButton)
        self.statusbar.addPermanentWidget(self.userLogsButton)

        # Connection status timer
        self.timerStatus = QTimer(self)
        self.timerStatus.timeout.connect(self.onMainTimerStatus)
        self.timerStatus.start(20000)

        self.enableButtons(False)

        # Connect the IPFS context signals
        self.app.ipfsCtx.ipfsConnectionReady.connect(self.onConnReady)
        self.app.ipfsCtx.ipfsRepositoryReady.connect(self.onRepoReady)
        self.app.ipfsCtx.pubsub.psMessageRx.connect(self.onPubsubRx)
        self.app.ipfsCtx.pubsub.psMessageTx.connect(self.onPubsubTx)
        self.app.ipfsCtx.profilesAvailable.connect(self.onProfilesList)
        self.app.ipfsCtx.profileChanged.connect(self.onProfileChanged)
        self.app.ipfsCtx.pinItemStatusChanged.connect(self.onPinStatusChanged)
        self.app.ipfsCtx.pinItemsCount.connect(self.onPinItemsCount)
        self.app.ipfsCtx.pinFinished.connect(self.onPinFinished)

        # Misc signals
        self.app.marksLocal.feedMarkAdded.connect(self.onFeedMarkAdded)
        self.app.systemTray.messageClicked.connect(self.onSystrayMsgClicked)

        # Application signals
        self.app.manualAvailable.connect(self.onManualAvailable)

        self.tabWidget.removeTab(0)

        previousGeo = self.app.settingsMgr.mainWindowGeometry
        if previousGeo:
            self.restoreGeometry(previousGeo)
        previousState = self.app.settingsMgr.mainWindowState
        if previousState:
            self.restoreState(previousState)

        self.hashmarksPage = None
        self.pinStatusTab = pin.PinStatusWidget(self)

        self.pinIconLoading = getIcon('pin-blue-loading.png')
        self.pinIconNormal = getIcon('pin-black.png')

        self.setCentralWidget(self.centralWidget)

    @property
    def app(self):
        return self._app

    @property
    def allTabs(self):
        return self._allTabs

    @property
    def searchBar(self):
        return self.centralWidget.searchBar

    def cycleTabs(self):
        curIndex = self.tabWidget.currentIndex()
        if curIndex + 1 < self.tabWidget.count():
            self.tabWidget.setCurrentIndex(curIndex + 1)
        else:
            self.tabWidget.setCurrentIndex(0)

    def onClearHistory(self):
        self.app.urlHistory.clear()

    def onHashmarkClicked(self, path, title):
        ipfsPath = IPFSPath(path, autoCidConv=True)

        if ipfsPath.valid:
            ensure(self.app.resourceOpener.open(ipfsPath))

    def onMainToolbarMoved(self, orientation):
        self.toolbarMain.lastPos = self.toolbarMain.pos()

        self.toolbarTools.setOrientation(orientation)
        self.qaToolbar.setOrientation(orientation)

        if self.toolbarMain.vertical:
            self.toolbarMain.emptySpace.setSizePolicy(
                QSizePolicy.Minimum, QSizePolicy.Expanding)
            self.qaToolbar.setMinimumHeight(
                self.toolbarMain.height() / 3)
            self.qaToolbar.setMinimumWidth(32)
        elif self.toolbarMain.horizontal:
            self.qaToolbar.setMinimumWidth(
                self.width() / 3)
            self.qaToolbar.setMinimumHeight(32)
            self.toolbarMain.emptySpace.setSizePolicy(QSizePolicy.Expanding,
                                                      QSizePolicy.Minimum)

    def onOpenEventLog(self):
        self.addEventLogTab(current=True)

    def onShowUserLogs(self, checked):
        lowerPos = self.mapToGlobal(QPoint(self.width(), self.height()))

        popupPoint = QPoint(
            lowerPos.x() - self.logsPopupWindow.width() - 64,
            lowerPos.y() - self.logsPopupWindow.height() - 64
        )

        self.logsPopupWindow.move(popupPoint)
        self.logsBrowser.moveCursor(QTextCursor.End)
        self.logsPopupWindow.setVisible(checked)

    def onProfileEditDialog(self):
        runDialog(ProfileEditDialog, self.app.ipfsCtx.currentProfile,
                  title='Profile Edit dialog')

    def onProfileWebsiteUpdated(self):
        self.profileEditButton.setStyleSheet('''
            QToolButton {
                background-color: #B7CDC2;
            }
        ''')
        self.app.loop.call_later(
            3, self.profileEditButton.setStyleSheet,
            'QToolButton {}'
        )

    def onProfileInfoChanged(self, profile):
        # Regen website
        ensure(profile.userWebsite.update())

    @asyncify
    async def onProfileChanged(self, pName, profile):
        if not profile.initialized:
            return

        def hashmarksLoaded(nodeId, hmarks):
            try:
                self.hashmarksSearcher.register(nodeId, hmarks)
            except Exception as err:
                log.debug(str(err))

        profile.sharedHManager.hashmarksLoaded.connect(hashmarksLoaded)

        self.profileEditButton.setEnabled(False)
        await profile.userInfo.loaded
        self.profileEditButton.setEnabled(True)

        if profile.userWebsite:
            profile.userWebsite.websiteUpdated.connect(
                self.onProfileWebsiteUpdated
            )

        if not profile.userInfo.usernameSet:
            QToolTip.showText(
                self.profileEditButton.mapToGlobal(
                    QPoint(0, 16)),
                "Your profile's username is not set!",
                self.profileEditButton, QRect(0, 0, 0, 0), 1800
            )

        profile.userInfo.changed.connect(
            lambda: self.onProfileInfoChanged(profile)
        )

        for action in self.profilesActionGroup.actions():
            if action.data() == pName:
                action.setChecked(True)
                # Refresh the file manager
                filesM = self.findTabFileManager()
                if filesM:
                    filesM.setupModel()
                    filesM.pathSelectorDefault()

    def onProfileViewHomepage(self):
        self.addBrowserTab().browseFsPath(os.path.join(
            joinIpns(self.app.ipfsCtx.currentProfile.keyRootId), 'index.html'))

    def onProfilesList(self, pList):
        currentList = [action.data() for action in
                       self.profilesActionGroup.actions()]

        for pName in pList:
            if pName in currentList:
                continue

            action = QAction(self.profilesActionGroup,
                             checkable=True, text=pName)
            action.setData(pName)

        for action in self.profilesActionGroup.actions():
            self.menuUserProfile.addAction(action)

    def onRepoReady(self):
        self.enableButtons()

        self.browseButton.normalIcon()

        ensure(self.displayConnectionInfo())
        ensure(self.qaToolbar.init())
        ensure(self.hashmarkMgrButton.updateIcons())
        ensure(self.app.marksLocal.pyramidsInit())
        ensure(self.app.sqliteDb.feeds.start())

        if self.app.enableOrbital and self.app.ipfsCtx.orbitConnector is None:
            self.app.ipfsCtx.orbitConnector = GalacteekOrbitConnector(
                orbitDataPath=self.app.orbitDataLocation,
                servicePort=self.app.settingsMgr.getSetting(
                    CFG_SECTION_ORBITDB,
                    CFG_KEY_CONNECTOR_LISTENPORT
                )
            )
            ensure(self.orbitStart())

    @ipfsOp
    async def orbitStart(self, ipfsop):
        resp = await self.app.ipfsCtx.orbitConnector.start()

        if resp:
            orbitIcon = QPushButton()
            orbitIcon.setIcon(getIcon('orbitdb.png'))
            orbitIcon.setToolTip('OrbitDB: connected')
            self.statusbar.addPermanentWidget(orbitIcon)

            self.orbitManager = DatabasesManager(
                self.app.ipfsCtx.orbitConnector, self)
            self.toolbarTools.addWidget(self.orbitManager.button)

            await ipfsop.ctx.currentProfile.orbitalSetup(
                self.app.ipfsCtx.orbitConnector)

    def onSystrayMsgClicked(self):
        # Open last shown mark in the systray
        if self._lastFeedMark is not None:
            self.addBrowserTab().browseFsPath(self._lastFeedMark.path)

    def onFeedMarkAdded(self, feedname, mark):
        self._lastFeedMark = mark

        metadata = mark.markData.get('metadata', None)
        if not metadata:
            return

        message = """New entry in IPNS feed {feed}: {title}.
                 Click the message to open it""".format(
            feed=feedname,
            title=metadata['title'] if metadata['title'] else 'No title'
        )

        self.app.systemTrayMessage('IPNS', message, timeout=5000)

    def onIpfsInfos(self):
        dlg = IPFSInfosDialog(self.app)
        dlg.setWindowTitle(iIpfsInfos())
        self.app.task(dlg.loadInfos)
        dlg.exec_()

    def onConnReady(self):
        pass

    def onPubsubRx(self):
        now = QDateTime.currentDateTime()
        self.pubsubStatusButton.setIcon(getIcon('network-transmit.png'))
        self.pubsubStatusButton.setToolTip(
            'Pubsub: last message received {}'.format(now.toString()))

    def onPubsubTx(self):
        pass

    def showPinningStatusWidget(self):
        name = self.tabnPinning
        ft = self.findTabWithName(name)
        if ft:
            return self.tabWidget.setCurrentWidget(ft)

        tab = self.pinStatusTab
        self.registerTab(tab, name, current=True,
                         icon=getIcon('pin-zoom.png'))

    def onPinItemsCount(self, count):
        statusMsg = iItemsInPinningQueue(count)

        if count > 0:
            self.pinningStatusButton.setIcon(self.pinIconLoading)
        else:
            self.pinningStatusButton.setIcon(self.pinIconNormal)

        self.pinningStatusButton.setToolTip(statusMsg)
        self.pinningStatusButton.setStatusTip(statusMsg)

    def onPinFinished(self, path):
        pass

    def onPinStatusChanged(self, qname, path, status):
        pass

    def onManualAvailable(self, langCode, entry):
        lang = iLangEnglish() if langCode == 'en' else iUnknown()
        self.menuManual.addAction(lang, functools.partial(
                                  self.onOpenManual, langCode))

    def onOpenManual(self, lang):
        entry = self.app.manuals.getManualEntry(lang)
        if entry:
            self.addBrowserTab().browseIpfsHash(entry['Hash'])

    def enableButtons(self, flag=True):
        for btn in [
                self.clipboardManager,
                self.browseButton,
                self.fileManagerButton,
                self.peersButton,
                self.textEditorButton,
                self.atomButton,
                self.hashmarkMgrButton,
                self.hashmarksSearcher,
                self.profileEditButton]:
            btn.setEnabled(flag)

    def statusMessage(self, msg):
        self.lastLogLabel.setText(msg)
        self.lastLogLabel.setToolTip(msg)

    def registerTab(self, tab, name, icon=None, current=False,
                    tooltip=None):
        idx = None
        if icon:
            idx = self.tabWidget.addTab(tab, icon, name)
        else:
            idx = self.tabWidget.addTab(tab, name)

        self._allTabs.append(tab)

        if current is True:
            self.tabWidget.setCurrentWidget(tab)
            tab.setFocus(Qt.OtherFocusReason)

        if tooltip and idx:
            self.tabWidget.setTabToolTip(idx, tooltip)

    def findTabFileManager(self):
        return self.findTabWithName(self.tabnFManager)

    def findTabIndex(self, w):
        return self.tabWidget.indexOf(w)

    def findTabWithName(self, name):
        for idx in range(0, self.tabWidget.count()):
            tName = self.tabWidget.tabText(idx)

            if tName.strip() == name.strip():
                return self.tabWidget.widget(idx)

    def removeTabFromWidget(self, w):
        idx = self.tabWidget.indexOf(w)
        if idx:
            self.tabWidget.removeTab(idx)

    def onSettings(self):
        runDialog(settings.SettingsDialog, self.app)

    def onCloseAllTabs(self):
        self.tabWidget.clear()

    def onToggledPinAllGlobal(self, checked):
        self.pinAllGlobalChecked = checked

    def onAboutGalacteek(self):
        runDialog(AboutDialog, iAbout())

    def setConnectionInfoMessage(self, msg):
        self.app.systemTray.setToolTip('{app}: {msg}'.format(
            app=GALACTEEK_NAME, msg=msg))
        self.ipfsStatusCube.setToolTip(msg)

    @ipfsOp
    async def displayConnectionInfo(self, ipfsop):
        try:
            info = await ipfsop.client.core.id()
        except BaseException:
            self.setConnectionInfoMessage(iErrNoCx())
            return

        nodeId = info.get('ID', iUnknown())
        nodeAgent = info.get('AgentVersion', iUnknownAgent())

        # Get IPFS peers list
        peers = await ipfsop.peersList()
        if not peers:
            self.setConnectionInfoMessage(iCxButNoPeers(nodeId, nodeAgent))
            self.ipfsStatusCube.clip.setSpeed(0)
            return

        peersCount = len(peers)

        # TODO: compute something more precise, probably based on the
        # swarm's high/low config
        connQuality = int((peersCount * 100) / 1000)

        self.setConnectionInfoMessage(
            iConnectStatus(nodeId, nodeAgent, peersCount))
        self.ipfsStatusCube.clip.setSpeed(5 + min(connQuality, 80))

    def onMainTimerStatus(self):
        ensure(self.displayConnectionInfo())

    def keyPressEvent(self, event):
        modifiers = event.modifiers()

        if modifiers & Qt.ControlModifier:
            if event.key() == Qt.Key_W:
                idx = self.tabWidget.currentIndex()
                self.onTabCloseRequest(idx)

        super(MainWindow, self).keyPressEvent(event)

    def explore(self, cid):
        ipfsPath = IPFSPath(cid, autoCidConv=True)
        if ipfsPath.valid:
            ensure(self.exploreIpfsPath(ipfsPath))
        else:
            messageBox(iInvalidInput())

    @ipfsOp
    async def exploreIpfsPath(self, ipfsop, ipfsPath):
        cid = await ipfsPath.resolve(ipfsop)
        if not cid:
            return messageBox(iCannotResolve(str(ipfsPath)))

        path = IPFSPath(cid, autoCidConv=True)
        tabName = path.shortRepr()

        tooltip = 'CID explorer: {0}'.format(cid)
        view = ipfsview.IPFSHashExplorerStack(self, cid)
        self.registerTab(view, tabName, current=True,
                         icon=getIcon('hash.png'), tooltip=tooltip)

    def addMediaPlayerTab(self):
        name = self.tabnMediaPlayer
        ft = self.findTabWithName(name)
        if ft:
            return ft
        tab = mediaplayer.MediaPlayerTab(self)

        if tab.playerAvailable():
            self.registerTab(tab, name, icon=getIcon('multimedia.png'),
                             current=True)
            return tab
        else:
            messageBox(mediaplayer.iPlayerUnavailable())

    def mediaPlayerQueue(self, path, mediaName=None):
        tab = self.addMediaPlayerTab()
        if tab:
            tab.queueFromPath(path, mediaName=mediaName)

    def mediaPlayerPlay(self, path, mediaName=None):
        tab = self.addMediaPlayerTab()
        if tab:
            tab.playFromPath(path, mediaName=mediaName)

    def onTabCloseRequest(self, idx):
        tab = self.tabWidget.widget(idx)

        if tab not in self.allTabs:
            return False

        if tab.onClose() is True:
            self.tabWidget.removeTab(idx)
            self.allTabs.remove(tab)
            del tab

    def addEditorTab(self, path=None, editing=True):
        tab = TextEditorTab(editing=editing, parent=self)

        if isinstance(path, IPFSPath) and path.valid:
            tab.editor.display(path)
            tab.setToolTip(str(path))

        self.registerTab(tab, iTextEditor(),
                         icon=getIcon('text-editor.png'), current=True)

    def addIpfsSearchView(self):
        tab = ipfssearch.IPFSSearchTab(self)
        self.registerTab(tab, iIpfsSearch(), current=True,
                         icon=getIcon('search-engine.png'))
        tab.view.browser.setFocus(Qt.OtherFocusReason)

    def onOpenMediaPlayer(self):
        self.addMediaPlayerTab()

    def onOpenBrowserTabClicked(self, pinBrowsed=False):
        self.addBrowserTab(pinBrowsed=pinBrowsed)

    def onWriteNewDocumentClicked(self):
        w = textedit.AddDocumentWidget(self, parent=self.tabWidget)
        self.registerTab(w, 'New document', current=True)

    def onPeersMgrClicked(self):
        self.showPeersMgr(current=True)

    def onFileManagerClicked(self):
        name = self.tabnFManager

        icon = getIcon('folder-open.png')
        ft = self.findTabWithName(name)
        if ft:
            ft.fileManager.updateTree()
            return self.tabWidget.setCurrentWidget(ft)

        fileManagerTab = files.FileManagerTab(
            self.tabWidget, fileManager=self.fileManagerWidget)
        self.registerTab(fileManagerTab, name, current=True, icon=icon)
        fileManagerTab.fileManager.updateTree()

    def onIpfsKeysClicked(self):
        name = self.tabnKeys
        ft = self.findTabWithName(name)
        if ft:
            return self.tabWidget.setCurrentWidget(ft)

        keysTab = keys.KeysTab(self)
        self.registerTab(keysTab, name, current=True)

    def onHelpDonate(self):
        bcAddress = '3HSsNcwzkiWGu6wB18BC6D37JHExpxZvyS'
        runDialog(DonateDialog, bcAddress)

    def addBrowserTab(self, label='No page loaded', pinBrowsed=False,
                      current=True):
        icon = getIconIpfsIce()
        tab = browser.BrowserTab(self, pinBrowsed=pinBrowsed)
        self.registerTab(tab, label, icon=icon, current=current)

        if self.app.settingsMgr.isTrue(CFG_SECTION_BROWSER, CFG_KEY_GOTOHOME):
            tab.loadHomePage()
        else:
            tab.focusUrlZone()

        return tab

    def addEventLogTab(self, current=False):
        self.registerTab(eventlog.EventLogWidget(self), iEventLog(),
                         current=current)

    def showPeersMgr(self, current=False):
        # Peers mgr
        name = iPeers()

        ft = self.findTabWithName(name)
        if ft:
            return self.tabWidget.setCurrentWidget(ft)

        pMgr = peers.PeersManager(self, self.app.peersTracker)
        self.registerTab(pMgr, name, current=current)

    def quit(self):
        # Qt and application exit
        self.saveUiSettings()
        ensure(self.app.exitApp())

    def saveUiSettings(self):
        self.app.settingsMgr.setSetting(
            CFG_SECTION_UI,
            CFG_KEY_MAINWINDOW_GEOMETRY,
            self.saveGeometry())
        self.app.settingsMgr.setSetting(
            CFG_SECTION_UI,
            CFG_KEY_MAINWINDOW_STATE,
            self.saveState())

        self.app.settingsMgr.setSetting(
            CFG_SECTION_UI,
            CFG_KEY_BROWSER_AUTOPIN,
            self.pinAllGlobalButton.isChecked())

    def closeEvent(self, event):
        event.ignore()
        self.showMinimized()

    def addHashmarksTab(self):
        ft = self.findTabWithName(self.tabnHashmarks)
        if ft:
            return self.tabWidget.setCurrentWidget(ft)

        tab = WebTab(self.tabWidget)

        if self.hashmarksPage is None:
            self.hashmarksPage = HashmarksPage(self.app.marksLocal,
                                               self.app.marksNetwork,
                                               parent=tab)
        hview = DWebView(page=self.hashmarksPage, parent=tab)
        tab.attach(hview)

        self.registerTab(tab, iHashmarks(),
                         icon=getIcon('hashmarks.png'), current=True)

    def onOpenChatWidget(self):
        tab = self.findTabWithName(self.tabnChat)
        if tab:
            tab.focusMessage()
            return self.tabWidget.setCurrentWidget(ft)

        self.chatRoomWidget.focusMessage()

        self.registerTab(self.chatRoomWidget, self.tabnChat,
                         icon=getIcon('chat.png'), current=True)

    def onShowAtomFeeds(self):
        ft = self.findTabWithName(self.tabnFeeds)
        if ft:
            return self.tabWidget.setCurrentWidget(ft)

        tab = AtomFeedsViewTab(self, view=self.atomFeedsViewWidget)
        self.registerTab(tab, self.tabnFeeds,
                         icon=getIcon('atom-feed.png'), current=True)

    def onIpfsObjectServed(self, ipfsPath, cType, reqTime):
        # TODO
        # Called when an object was served by the native IPFS scheme handler
        pass



import functools
import os.path
import os
import aioipfs

from PyQt5.QtWebEngineWidgets import QWebEngineDownloadItem
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import QObject

from galacteek import log
from galacteek import logUser
from galacteek import ensure
from galacteek.ipfs import ipfsOp
from galacteek.ipfs.crawl import runTitleParser
from galacteek.core.asynclib import asyncReadFile
from galacteek.core.asynclib import asyncWriteFile

from .helpers import questionBox
from ..appsettings import *
from .i18n import iUnknown


def iFinishedDownload(filename):
    return QCoreApplication.translate(
        'Galacteek', 'Finished downloading file {0}').format(filename)


def iStartingDownload(filename):
    return QCoreApplication.translate(
        'Galacteek', 'Downloading file {0} ..').format(filename)


def iPageSaved(title):
    return QCoreApplication.translate(
        'Galacteek',
        'Web page {0} was saved').format(title)


class DownloadsManager(QObject):
    def __init__(self, app):
        super(DownloadsManager, self).__init__(app)
        self.app = app

    def onDownloadRequested(self, downItem):
        if not downItem:
            return

        downloadsLoc = self.app.settingsMgr.eGet(S_DOWNLOADS_PATH)

        def progress(received, total):
            pass

        def finished(item):
            filename = item.path() or iUnknown()
            self.app.systemTrayMessage(
                'Galacteek', iFinishedDownload(filename))

        downPath = downItem.path()
        name = os.path.basename(downPath)
        autoDownloadFor = ['qwe_download']

        if name in autoDownloadFor:
            downItem.accept()
            return

        if downPath.startswith(self.app.tempDirWeb):
            downItem.finished.connect(
                functools.partial(self.pageSaved, downItem))
            downItem.accept()
            return

        reply = questionBox('Download', 'Download {} ?'.format(name))
        if reply is not True:
            downItem.cancel()
            return

        self.app.systemTrayMessage('Galacteek', iStartingDownload(name))

        downItem.setPath(os.path.join(downloadsLoc, name))
        downItem.finished.connect(functools.partial(finished, downItem))
        downItem.downloadProgress.connect(progress)
        downItem.accept()

    def pageSaved(self, downItem):
        saveFormat = downItem.savePageFormat()

        if saveFormat == QWebEngineDownloadItem.CompleteHtmlSaveFormat:
            ensure(self.pageSavedComplete(downItem))

    @ipfsOp
    async def pageSavedComplete(self, ipfsop, downItem):
        curProfile = ipfsop.ctx.currentProfile

        path = downItem.path()
        basedir = os.path.dirname(path)

        data = await asyncReadFile(path)
        if data is None:
            return

        # Rewrite links that have the gateway origin
        try:
            decoded = data.decode()
            params = self.app.getIpfsConnectionParams()
            replaced = decoded.replace(str(params.gatewayUrl), 'dweb:')
        except:
            # No rewrite
            pass
        else:
            await asyncWriteFile(path, replaced.encode())

        title = runTitleParser(data)
        if title is None:
            title = iUnknown()

        try:
            entry = await ipfsop.addPath(basedir)
        except aioipfs.APIError as err:
            log.debug('Cannot import saved page: {}'.format(err.message))
            return
        else:
            if entry is None:
                return

            cid = entry.get('Hash')
            logUser.debug('Saved webpage to {}'.format(cid))

            if curProfile and cid:
                curProfile.webPageSaved.emit(entry, title)

            self.app.systemTrayMessage('Downloads', iPageSaved(title))



from PyQt5.QtWidgets import QTreeView
from PyQt5.QtWidgets import QApplication

from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import Qt

from PyQt5.QtGui import QStandardItemModel
from PyQt5.QtGui import QBrush
from PyQt5.QtGui import QColor

from galacteek import ensure
from galacteek.core.modelhelpers import UneditableItem

from .i18n import iUnknown
from .i18n import iHashmarks


class URLHistory(QObject):
    historyConfigChanged = pyqtSignal(bool)

    def __init__(self, db, enabled=False, parent=None):
        super(URLHistory, self).__init__(parent)
        self.db = db
        self.enabled = enabled
        self.historyConfigChanged.connect(self.onConfig)

    def onConfig(self, enableHistory):
        self.enabled = enableHistory

    def record(self, url, title):
        if self.enabled:
            ensure(self.db.historyRecord(url, title))

    def clear(self):
        ensure(self.db.historyClear())

    async def match(self, input):
        return await self.db.historySearch(input)


class HistoryMatchesWidget(QTreeView):
    historyItemSelected = pyqtSignal(str)
    collapsed = pyqtSignal()

    def __init__(self, parent=None):
        super(HistoryMatchesWidget, self).__init__(parent)

        self.app = QApplication.instance()
        self.setObjectName('historySearchResults')
        self.clicked.connect(self.onItemActivated)

        self.model = QStandardItemModel()
        self.setModel(self.model)
        self.setHeaderHidden(True)

    def onItemActivated(self, idx):
        idxUrl = self.model.index(idx.row(), 1, idx.parent())
        data = self.model.data(idxUrl)

        if isinstance(data, str) and data:
            self.historyItemSelected.emit(data)

    def showMatches(self, marks, hMatches):
        self.model.clear()
        brush = QBrush(QColor('lightgrey'))

        mItem = UneditableItem(iHashmarks())
        mItem.setBackground(brush)
        mItemE = UneditableItem('')
        mItemE.setBackground(brush)
        self.model.invisibleRootItem().appendRow([mItem, mItemE])

        for match in marks:
            title = match['title'][0:64] if match['title'] else iUnknown()

            itemT = UneditableItem(title)
            item = UneditableItem(match['url'])
            item.setToolTip(match['url'])
            item.setData(match['url'], Qt.EditRole)

            mItem.appendRow([itemT, item])

        hItem = UneditableItem('History items')
        hItemE = UneditableItem('')
        hItem.setBackground(brush)
        hItemE.setBackground(brush)
        self.model.invisibleRootItem().appendRow([hItem, hItemE])

        for match in hMatches:
            title = match['title'][0:64] if match['title'] else iUnknown()
            itemT = UneditableItem(title)

            item = UneditableItem(match['url'])
            item.setToolTip(match['url'])
            item.setData(match['url'], Qt.EditRole)

            hItem.appendRow([itemT, item])

        self.expandAll()
        self.resizeColumnToContents(0)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return:
            # The 'activated' signal does not seem to have the same
            # behavior across platforms so we handle Return manually here
            curIdx = self.currentIndex()
            if curIdx.isValid():
                self.onItemActivated(curIdx)
        elif event.key() == Qt.Key_Escape:
            self.hide()
            self.collapsed.emit()
        else:
            super(HistoryMatchesWidget, self).keyPressEvent(event)



import importlib
from galacteek import ensure

from galacteek import log
from galacteek.core.objects import GObject
from galacteek.ipdapps import availableDapps
from galacteek.ipdapps import dappsPkg
from galacteek.smartcontracts import getContractByName


class DappsRegistry(GObject):
    def __init__(self, ethController, parent=None):
        super().__init__(parent)
        self._dapps = {}
        self.ethCtrl = ethController

    async def loadDappFromAddr(self, dappName, contractName, address):
        lContract = getContractByName(contractName)
        if not lContract:
            return

        def load(_lContract, address):
            # Blocking call ?
            w3Contract = _lContract.web3Contract(self.web3)
            return w3Contract(address)

        contract = await self._e(load, lContract, address)
        if contract:
            return lContract.module.contractOperator(self, contract,
                                                     address, parent=self)
        else:
            log.debug('Contract invalid: {}'.format(address))

    async def loadDapp(self, dappName, **params):
        dapps = availableDapps()
        dappPath = dapps.get(dappName, None)

        if not dappPath:
            return None

        modname = '{0}.{1}'.format(dappsPkg, dappName)

        try:
            module = importlib.import_module(modname)
        except Exception as err:
            print(str(err))
            return None

        if module:
            dapp = module.Dapp(
                self.ethCtrl,
                dappName,
                dappPath,
                initModule=module,
                parent=self,
                **params
            )
            dapp.pkgContentsChanged.connect(self.onDappPkgChanged)
            await dapp.registerSchemeHandlers()
            await dapp.init()
            self._dapps[dappName] = dapp
            return dapp

    def onDappPkgChanged(self, dappName):
        dapp = self._dapps.get(dappName)
        if dapp:
            dapp.pkgContentsChanged.disconnect(self.onDappPkgChanged)
            dapp.reimportModule()
            dapp.stop()
            del self._dapps[dappName]

            ensure(self.loadDapp(dappName))
